# Import key components that should be publicly accessible
from judgeval.clients import client, together_client
from judgeval.judgment_client import JudgmentClient

__all__ = [
    # Clients
    'client',
    'together_client',
    'JudgmentClient',
]



import ast
import csv
import datetime
import json
import os
import yaml
from dataclasses import dataclass, field
from typing import List, Union, Literal

from judgeval.data import Example
from judgeval.common.logger import debug, error, warning, info

@dataclass
class EvalDataset:
    examples: List[Example]
    _alias: Union[str, None] = field(default=None)
    _id: Union[str, None] = field(default=None)
    judgment_api_key: str = field(default="")
    organization_id: str = field(default="")
    def __init__(self, 
                 judgment_api_key: str = os.getenv("JUDGMENT_API_KEY"),  
                 organization_id: str = os.getenv("JUDGMENT_ORG_ID"),
                 examples: List[Example] = [],
                 ):
        debug(f"Initializing EvalDataset with {len(examples)} examples")
        if not judgment_api_key:
            warning("No judgment_api_key provided")
        self.examples = examples
        self._alias = None
        self._id = None
        self.judgment_api_key = judgment_api_key
        self.organization_id = organization_id

    def add_from_json(self, file_path: str) -> None:
        debug(f"Loading dataset from JSON file: {file_path}")
        """
        Adds examples from a JSON file.

        The format of the JSON file is expected to be a dictionary with one key: "examples". 
        The value of the key is a list of dictionaries, where each dictionary represents an example.

        The JSON file is expected to have the following format:
        {
        "examples": [
            {
                "input": "test input",
                "actual_output": "test output",
                "expected_output": "expected output",
                "context": [
                    "context1",
                    "context2"
                ],
                "retrieval_context": [
                    "retrieval1"
                ],
                "additional_metadata": {
                    "key": "value"
                },
                "tools_called": [
                    "tool1"
                ],
                "expected_tools": [
                    "tool1",
                    "tool2"
                ],
                "name": "test example",
                "example_id": null,
                "timestamp": "20241230_160117",
                "trace_id": "123"
            }
            ]
        }
        """
        try:
            with open(file_path, "r") as file:
                payload = json.load(file)
                examples = payload.get("examples", [])
        except FileNotFoundError:
            error(f"JSON file not found: {file_path}")
            raise FileNotFoundError(f"The file {file_path} was not found.")
        except json.JSONDecodeError:
            error(f"Invalid JSON file: {file_path}")
            raise ValueError(f"The file {file_path} is not a valid JSON file.")

        info(f"Added {len(examples)} examples from JSON")
        new_examples = [Example(**e) for e in examples]
        for e in new_examples:
            self.add_example(e)
  
    def add_from_csv(
        self, 
        file_path: str,
        header_mapping: dict,
        primary_delimiter: str = ",",
        secondary_delimiter: str = ";"
        ) -> None:
        """
        Add Examples from a CSV file.

        Args:
            file_path (str): Path to the CSV file
            header_mapping (dict): Dictionary mapping Example headers to custom headers
            primary_delimiter (str, optional): Main delimiter used in CSV file. Defaults to ","
            secondary_delimiter (str, optional): Secondary delimiter for list fields. Defaults to ";"
        """
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Please install pandas to use this method. 'pip install pandas'"
            )
        
        # Pandas naturally reads numbers in data files as ints, not strings (can lead to unexpected behavior)
        df = pd.read_csv(file_path, dtype={'trace_id': str}, sep=primary_delimiter)
        """
        The user should pass in a dict mapping from Judgment Example headers to their custom defined headers.
        Available headers for Example objects are as follows:

        "input", "actual_output", "expected_output", "context", \
        "retrieval_context", "additional_metadata", "tools_called", \
        "expected_tools", "name", "comments", "source_file", "example", \
        "trace_id"

        We want to collect the examples separately which can
        be determined by the "example" column. If the value is True, then it is an
        example, and we expect the `input` and `actual_output` fields to be non-null.

        We also assume that if there are multiple retrieval contexts, contexts, or tools called, they are separated by semicolons.
        This can be adjusted using the `secondary_delimiter` parameter.
        """
        examples = []
        
        def process_csv_row(value, header):
            """
            Maps a singular value in the CSV file to the appropriate type based on the header.
            If value exists and can be split into type List[*], we will split upon the user's provided secondary delimiter.
            """
            # check that the CSV value is not null for entry
            null_replacement = dict() if header == 'additional_metadata' else None
            if pd.isna(value) or value == '':
                return null_replacement
            try:
                value = ast.literal_eval(value) if header == 'additional_metadata' else str(value)
            except (ValueError, SyntaxError):
                value = str(value)
            if header in ["context", "retrieval_context", "tools_called", "expected_tools"]:
                # attempt to split the value by the secondary delimiter
                value = value.split(secondary_delimiter)
                    
            return value
            
        for _, row in df.iterrows():
            data = {
                header: process_csv_row(
                    row[header_mapping[header]], header
                )
                for header in header_mapping
            }
            if "example" in header_mapping and row[header_mapping["example"]]:
                if "name" in header_mapping:
                    data["name"] = row[header_mapping["name"]] if pd.notna(row[header_mapping["name"]]) else None
                # every Example has `input` and `actual_output` fields
                if data["input"] is not None and data["actual_output"] is not None:
                    e = Example(**data)
                    examples.append(e)
                else:
                    raise ValueError("Every example must have an 'input' and 'actual_output' field.")

        for e in examples:
            self.add_example(e)

    def add_from_yaml(self, file_path: str) -> None:
        debug(f"Loading dataset from YAML file: {file_path}")
        """
        Adds examples from a YAML file.

        The format of the YAML file is expected to be a dictionary with one key: "examples". 
        The value of the key is a list of dictionaries, where each dictionary represents an example.

        The YAML file is expected to have the following format:
        examples:
          - input: "test input"
            actual_output: "test output"
            expected_output: "expected output"
            context:
              - "context1"
              - "context2"
            retrieval_context:
              - "retrieval1"
            additional_metadata:
              key: "value"
            tools_called:
              - "tool1"
            expected_tools:
              - "tool1"
              - "tool2"
            name: "test example"
            example_id: null
            timestamp: "20241230_160117"
            trace_id: "123"
        """
        try:
            with open(file_path, "r") as file:
                payload = yaml.safe_load(file)
                if payload is None:
                    raise ValueError("The YAML file is empty.")
                examples = payload.get("examples", [])
        except FileNotFoundError:
            error(f"YAML file not found: {file_path}")
            raise FileNotFoundError(f"The file {file_path} was not found.")
        except yaml.YAMLError:
            error(f"Invalid YAML file: {file_path}")
            raise ValueError(f"The file {file_path} is not a valid YAML file.")

        info(f"Added {len(examples)} examples from YAML")
        new_examples = [Example(**e) for e in examples]
        for e in new_examples:
            self.add_example(e)

    def add_example(self, e: Example) -> None:
        self.examples = self.examples + [e]
        # TODO if we need to add rank, then we need to do it here
    
    def save_as(self, file_type: Literal["json", "csv", "yaml"], dir_path: str, save_name: str = None) -> None:
        """
        Saves the dataset as a file. Save only the examples.

        Args:
            file_type (Literal["json", "csv"]): The file type to save the dataset as.
            dir_path (str): The directory path to save the file to.
            save_name (str, optional): The name of the file to save. Defaults to None.
        """
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        file_name = datetime.datetime.now().strftime("%Y%m%d_%H%M%S") if save_name is None else save_name
        complete_path = os.path.join(dir_path, f"{file_name}.{file_type}")
        if file_type == "json":
            with open(complete_path, "w") as file:
                json.dump(
                    {
                        "examples": [e.to_dict() for e in self.examples],
                    },
                    file,
                    indent=4,
                )
        elif file_type == "csv":
            with open(complete_path, "w", newline="") as file:
                writer = csv.writer(file)
                writer.writerow([
                    "input", "actual_output", "expected_output", "context", \
                    "retrieval_context", "additional_metadata", "tools_called", \
                    "expected_tools", "name", "comments", "source_file", "example", \
                    "trace_id"
                ])
                for e in self.examples:
                    writer.writerow(
                        [
                            e.input,
                            e.actual_output,
                            e.expected_output,
                            ";".join(e.context),
                            ";".join(e.retrieval_context),
                            e.additional_metadata,
                            ";".join(e.tools_called),
                            ";".join(e.expected_tools),
                            e.name,
                            None,  # Example does not have comments
                            None,  # Example does not have source file
                            True,  # Adding an Example
                            e.trace_id
                        ]
                    )
                
                
        elif file_type == "yaml":
            with open(complete_path, "w") as file:
                yaml_data = {
                    "examples": [
                        {
                            "input": e.input,
                            "actual_output": e.actual_output,
                            "expected_output": e.expected_output,
                            "context": e.context,
                            "retrieval_context": e.retrieval_context,
                            "additional_metadata": e.additional_metadata,
                            "tools_called": e.tools_called,
                            "expected_tools": e.expected_tools,
                            "name": e.name,
                            "comments": None,  # Example does not have comments
                            "source_file": None,  # Example does not have source file
                            "example": True,  # Adding an Example
                            "trace_id": e.trace_id
                        }
                        for e in self.examples
                    ],
                }
                yaml.dump(yaml_data, file, default_flow_style=False)
        else:
            ACCEPTABLE_FILE_TYPES = ["json", "csv", "yaml"]
            raise TypeError(f"Invalid file type: {file_type}. Please choose from {ACCEPTABLE_FILE_TYPES}")
        
    def __iter__(self):
        return iter(self.examples)
    
    def __len__(self):
        return len(self.examples)
    
    def __str__(self):
        return (
            f"{self.__class__.__name__}("
            f"examples={self.examples}, "
            f"_alias={self._alias}, "
            f"_id={self._id}"
            f")"
        )


"""
Tracing system for judgeval that allows for function tracing using decorators.
"""
# Standard library imports
import asyncio
import functools
import inspect
import json
import os
import time
import uuid
import warnings
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from http import HTTPStatus
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, TypeAlias, Union
from rich import print as rprint

# Third-party imports
import pika
import requests
from litellm import cost_per_token
from pydantic import BaseModel
from rich import print as rprint
from openai import OpenAI
from together import Together
from anthropic import Anthropic

# Local application/library-specific imports
from judgeval.constants import (
    JUDGMENT_TRACES_SAVE_API_URL,
    JUDGMENT_TRACES_FETCH_API_URL,
    RABBITMQ_HOST,
    RABBITMQ_PORT,
    RABBITMQ_QUEUE,
    JUDGMENT_TRACES_DELETE_API_URL,
    JUDGMENT_PROJECT_DELETE_API_URL,
    JUDGMENT_TRACES_ADD_TO_EVAL_QUEUE_API_URL
)
from judgeval.judgment_client import JudgmentClient
from judgeval.data import Example
from judgeval.scorers import APIJudgmentScorer, JudgevalScorer, ScorerWrapper
from judgeval.rules import Rule
from judgeval.evaluation_run import EvaluationRun
from judgeval.data.result import ScoringResult


# Define type aliases for better code readability and maintainability
ApiClient: TypeAlias = Union[OpenAI, Together, Anthropic]  # Supported API clients
TraceEntryType = Literal['enter', 'exit', 'output', 'input', 'evaluation']  # Valid trace entry types
SpanType = Literal['span', 'tool', 'llm', 'evaluation', 'chain']
@dataclass
class TraceEntry:
    """Represents a single trace entry with its visual representation.
    
    Visual representations:
    - enter: → (function entry)
    - exit: ← (function exit)
    - output: Output: (function return value)
    - input: Input: (function parameters)
    - evaluation: Evaluation: (evaluation results)
    """
    type: TraceEntryType
    function: str  # Name of the function being traced
    depth: int    # Indentation level for nested calls
    message: str  # Human-readable description
    timestamp: float  # Unix timestamp when entry was created
    duration: Optional[float] = None  # Time taken (for exit/evaluation entries)
    output: Any = None  # Function output value
    # Use field() for mutable defaults to avoid shared state issues
    inputs: dict = field(default_factory=dict)
    span_type: SpanType = "span"
    evaluation_runs: List[Optional[EvaluationRun]] = field(default=None)
    
    def print_entry(self):
        indent = "  " * self.depth
        if self.type == "enter":
            print(f"{indent}→ {self.function} (trace: {self.message})")
        elif self.type == "exit":
            print(f"{indent}← {self.function} ({self.duration:.3f}s)")
        elif self.type == "output":
            print(f"{indent}Output: {self.output}")
        elif self.type == "input":
            print(f"{indent}Input: {self.inputs}")
        elif self.type == "evaluation":
            for evaluation_run in self.evaluation_runs:
                print(f"{indent}Evaluation: {evaluation_run.model_dump()}")
    
    def _serialize_inputs(self) -> dict:
        """Helper method to serialize input data safely.
        
        Returns a dict with serializable versions of inputs, converting non-serializable
        objects to None with a warning.
        """
        serialized_inputs = {}
        for key, value in self.inputs.items():
            if isinstance(value, BaseModel):
                serialized_inputs[key] = value.model_dump()
            elif isinstance(value, (list, tuple)):
                # Handle lists/tuples of arguments
                serialized_inputs[key] = [
                    item.model_dump() if isinstance(item, BaseModel)
                    else None if not self._is_json_serializable(item)
                    else item
                    for item in value
                ]
            else:
                if self._is_json_serializable(value):
                    serialized_inputs[key] = value
                else:
                    serialized_inputs[key] = self.safe_stringify(value, self.function)
        return serialized_inputs

    def _is_json_serializable(self, obj: Any) -> bool:
        """Helper method to check if an object is JSON serializable."""
        try:
            json.dumps(obj)
            return True
        except (TypeError, OverflowError, ValueError):
            return False

    def safe_stringify(self, output, function_name):
        """
        Safely converts an object to a string or repr, handling serialization issues gracefully.
        """
        try:
            return str(output)
        except (TypeError, OverflowError, ValueError):
            pass
    
        try:
            return repr(output)
        except (TypeError, OverflowError, ValueError):
            pass
    
        warnings.warn(
            f"Output for function {function_name} is not JSON serializable and could not be converted to string. Setting to None."
        )
        return None

    def to_dict(self) -> dict:
        """Convert the trace entry to a dictionary format for storage/transmission."""
        return {
            "type": self.type,
            "function": self.function,
            "depth": self.depth,
            "message": self.message,
            "timestamp": self.timestamp,
            "duration": self.duration,
            "output": self._serialize_output(),
            "inputs": self._serialize_inputs(),
            "evaluation_runs": [evaluation_run.model_dump() for evaluation_run in self.evaluation_runs] if self.evaluation_runs else [],
            "span_type": self.span_type
        }

    def _serialize_output(self) -> Any:
        """Helper method to serialize output data safely.
        
        Handles special cases:
        - Pydantic models are converted using model_dump()
        - We try to serialize into JSON, then string, then the base representation (__repr__)
        - Non-serializable objects return None with a warning
        """
        
        if isinstance(self.output, BaseModel):
            return self.output.model_dump()
        
        try:
            # Try to serialize the output to verify it's JSON compatible
            json.dumps(self.output)
            return self.output
        except (TypeError, OverflowError, ValueError):
            return self.safe_stringify(self.output, self.function)
        

class TraceManagerClient:
    """
    Client for handling trace endpoints with the Judgment API
    

    Operations include:
    - Fetching a trace by id
    - Saving a trace
    - Deleting a trace
    """
    def __init__(self, judgment_api_key: str, organization_id: str):
        self.judgment_api_key = judgment_api_key
        self.organization_id = organization_id

    def fetch_trace(self, trace_id: str):
        """
        Fetch a trace by its id
        """
        response = requests.post(
            JUDGMENT_TRACES_FETCH_API_URL,
            json={
                "trace_id": trace_id,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.judgment_api_key}",
                "X-Organization-Id": self.organization_id
            },
            verify=True
        )

        if response.status_code != HTTPStatus.OK:
            raise ValueError(f"Failed to fetch traces: {response.text}")
        
        return response.json()

    def save_trace(self, trace_data: dict, empty_save: bool):
        """
        Saves a trace to the database

        Args:
            trace_data: The trace data to save
            empty_save: Whether to save an empty trace
            NOTE we save empty traces in order to properly handle async operations; we need something in the DB to associate the async results with
        """
        response = requests.post(
            JUDGMENT_TRACES_SAVE_API_URL,
            json=trace_data,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.judgment_api_key}",
                "X-Organization-Id": self.organization_id
            },
            verify=True
        )
        
        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(f"Failed to save trace data: Check your Trace name for conflicts, set overwrite=True to overwrite existing traces: {response.text}")
        elif response.status_code != HTTPStatus.OK:
            raise ValueError(f"Failed to save trace data: {response.text}")
        
        if not empty_save and "ui_results_url" in response.json():
            pretty_str = f"\n🔍 You can view your trace data here: [rgb(106,0,255)][link={response.json()['ui_results_url']}]View Trace[/link]\n"
            rprint(pretty_str)

    def delete_trace(self, trace_id: str):
        """
        Delete a trace from the database.
        """
        response = requests.delete(
            JUDGMENT_TRACES_DELETE_API_URL,
            json={
                "trace_ids": [trace_id],
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.judgment_api_key}",
                "X-Organization-Id": self.organization_id
            }
        )

        if response.status_code != HTTPStatus.OK:
            raise ValueError(f"Failed to delete trace: {response.text}")
        
        return response.json()
    
    def delete_traces(self, trace_ids: List[str]):
        """
        Delete a batch of traces from the database.
        """
        response = requests.delete(
            JUDGMENT_TRACES_DELETE_API_URL,
            json={
                "trace_ids": trace_ids,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.judgment_api_key}",
                "X-Organization-Id": self.organization_id
            }
        )

        if response.status_code != HTTPStatus.OK:
            raise ValueError(f"Failed to delete trace: {response.text}")
        
        return response.json()
    
    def delete_project(self, project_name: str):
        """
        Deletes a project from the server. Which also deletes all evaluations and traces associated with the project.
        """
        response = requests.delete(
            JUDGMENT_PROJECT_DELETE_API_URL,
            json={
                "project_name": project_name,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.judgment_api_key}",
                "X-Organization-Id": self.organization_id
            }
        )

        if response.status_code != HTTPStatus.OK:
            raise ValueError(f"Failed to delete traces: {response.text}")
            
        return response.json()


class TraceClient:
    """Client for managing a single trace context"""
    
    def __init__(
        self,
        tracer: Optional["Tracer"],
        trace_id: Optional[str] = None,
        name: str = "default",
        project_name: str = "default_project",
        overwrite: bool = False,
        rules: Optional[List[Rule]] = None,
        enable_monitoring: bool = True,
        enable_evaluations: bool = True
    ):
        self.name = name
        self.trace_id = trace_id or str(uuid.uuid4())
        self.project_name = project_name
        self.overwrite = overwrite
        self.tracer = tracer
        # Initialize rules with either provided rules or an empty list
        self.rules = rules or []
        self.enable_monitoring = enable_monitoring
        self.enable_evaluations = enable_evaluations
        
        self.client: JudgmentClient = tracer.client
        self.entries: List[TraceEntry] = []
        self.start_time = time.time()
        self.span_type = None
        self._current_span: Optional[TraceEntry] = None
        self.trace_manager_client = TraceManagerClient(tracer.api_key, tracer.organization_id)  # Manages DB operations for trace data
        self.visited_nodes = []  # Track nodes visited through langgraph_node spans
        self.executed_tools = []  # Track tools executed through tool spans
        self.executed_node_tools = []  # Track node:tool combinations
        
    @contextmanager
    def span(self, name: str, span_type: SpanType = "span"):
        """Context manager for creating a trace span"""
        start_time = time.time()
        
        # Record span entry
        self.add_entry(TraceEntry(
            type="enter",
            function=name,
            depth=self.tracer.depth,
            message=name,
            timestamp=start_time,
            span_type=span_type
        ))
        
        # Increment nested depth and set current span
        self.tracer.depth += 1
        prev_span = self._current_span
        self._current_span = name
        
        try:
            yield self
        finally:
            self.tracer.depth -= 1
            duration = time.time() - start_time
            
            # Record span exit
            self.add_entry(TraceEntry(
                type="exit",
                function=name,
                depth=self.tracer.depth,
                message=f"← {name}",
                timestamp=time.time(),
                duration=duration,
                span_type=span_type
            ))
            self._current_span = prev_span
            
    def async_evaluate(
        self,
        scorers: List[Union[APIJudgmentScorer, JudgevalScorer]],
        input: Optional[str] = None,
        actual_output: Optional[str] = None,
        expected_output: Optional[str] = None,
        context: Optional[List[str]] = None,
        retrieval_context: Optional[List[str]] = None,
        tools_called: Optional[List[str]] = None,
        expected_tools: Optional[List[str]] = None,
        additional_metadata: Optional[Dict[str, Any]] = None,
        model: Optional[str] = None,
        log_results: Optional[bool] = True
    ):
        if not self.enable_evaluations:
            return
        
        start_time = time.time()  # Record start time
        example = Example(
            input=input,
            actual_output=actual_output,
            expected_output=expected_output,
            context=context,
            retrieval_context=retrieval_context,
            tools_called=tools_called,
            expected_tools=expected_tools,
            additional_metadata=additional_metadata,
            trace_id=self.trace_id
        )
        loaded_rules = None
        if self.rules:
            loaded_rules = []
            for rule in self.rules:
                processed_conditions = []
                for condition in rule.conditions:
                    # Convert metric if it's a ScorerWrapper
                    try:
                        if isinstance(condition.metric, ScorerWrapper):
                            condition_copy = condition.model_copy()
                            condition_copy.metric = condition.metric.load_implementation(use_judgment=True)
                            processed_conditions.append(condition_copy)
                        else:
                            processed_conditions.append(condition)
                    except Exception as e:
                        warnings.warn(f"Failed to convert ScorerWrapper in rule '{rule.name}', condition metric '{condition.metric_name}': {str(e)}")
                        processed_conditions.append(condition)  # Keep original condition as fallback
                
                # Create new rule with processed conditions
                new_rule = rule.model_copy()
                new_rule.conditions = processed_conditions
                loaded_rules.append(new_rule)
        try:
            # Load appropriate implementations for all scorers
            loaded_scorers: List[Union[JudgevalScorer, APIJudgmentScorer]] = []
            for scorer in scorers:
                try:
                    if isinstance(scorer, ScorerWrapper):
                        loaded_scorers.append(scorer.load_implementation(use_judgment=True))
                    else:
                        loaded_scorers.append(scorer)
                except Exception as e:
                    warnings.warn(f"Failed to load implementation for scorer {scorer}: {str(e)}")
                    # Skip this scorer
            
            if not loaded_scorers:
                warnings.warn("No valid scorers available for evaluation")
                return
            
            # Prevent using JudgevalScorer with rules - only APIJudgmentScorer allowed with rules
            if loaded_rules and any(isinstance(scorer, JudgevalScorer) for scorer in loaded_scorers):
                raise ValueError("Cannot use Judgeval scorers, you can only use API scorers when using rules. Please either remove rules or use only APIJudgmentScorer types.")
            
        except Exception as e:
            warnings.warn(f"Failed to load scorers: {str(e)}")
            return
        
        # Combine the trace-level rules with any evaluation-specific rules)
        eval_run = EvaluationRun(
            organization_id=self.tracer.organization_id,
            log_results=log_results,
            project_name=self.project_name,
            eval_name=f"{self.name.capitalize()}-"
                f"{self._current_span}-"
                f"[{','.join(scorer.score_type.capitalize() for scorer in loaded_scorers)}]",
            examples=[example],
            scorers=loaded_scorers,
            model=model,
            metadata={},
            judgment_api_key=self.tracer.api_key,
            override=self.overwrite,
            rules=loaded_rules # Use the combined rules
        )
        
        self.add_eval_run(eval_run, start_time)  # Pass start_time to record_evaluation
            
    def add_eval_run(self, eval_run: EvaluationRun, start_time: float):
        """
        Add evaluation run data to the trace

        Args:
            eval_run (EvaluationRun): The evaluation run to add to the trace
            start_time (float): The start time of the evaluation run
        """
        if self._current_span:
            duration = time.time() - start_time  # Calculate duration from start_time
            
            prev_entry = self.entries[-1]
            
            # Select the last entry in the trace if it's an LLM call, otherwise use the current span
            self.add_entry(TraceEntry(
                type="evaluation",
                function=prev_entry.function if prev_entry.span_type == "llm" else self._current_span,
                depth=self.tracer.depth,
                message=f"Evaluation results for {self._current_span}",
                timestamp=time.time(),
                evaluation_runs=[eval_run],
                duration=duration,
                span_type="evaluation"
            ))

    def record_input(self, inputs: dict):
        """Record input parameters for the current span"""
        if self._current_span:
            self.add_entry(TraceEntry(
                type="input",
                function=self._current_span,
                depth=self.tracer.depth,
                message=f"Inputs to {self._current_span}",
                timestamp=time.time(),
                inputs=inputs,
                span_type=self.span_type
            ))

    async def _update_coroutine_output(self, entry: TraceEntry, coroutine: Any):
        """Helper method to update the output of a trace entry once the coroutine completes"""
        try:
            result = await coroutine
            entry.output = result
            return result
        except Exception as e:
            entry.output = f"Error: {str(e)}"
            raise

    def record_output(self, output: Any):
        """Record output for the current span"""
        if self._current_span:
            entry = TraceEntry(
                type="output",
                function=self._current_span,
                depth=self.tracer.depth,
                message=f"Output from {self._current_span}",
                timestamp=time.time(),
                output="<pending>" if inspect.iscoroutine(output) else output,
                span_type=self.span_type
            )
            self.add_entry(entry)
            
            if inspect.iscoroutine(output):
                # Create a task to update the output once the coroutine completes
                asyncio.create_task(self._update_coroutine_output(entry, output))

    def add_entry(self, entry: TraceEntry):
        """Add a trace entry to this trace context"""
        self.entries.append(entry)
        return self
        
    def print(self):
        """Print the complete trace with proper visual structure"""
        for entry in self.entries:
            entry.print_entry()
            
    def get_duration(self) -> float:
        """
        Get the total duration of this trace
        """
        return time.time() - self.start_time
    
    def condense_trace(self, entries: List[dict]) -> List[dict]:
        """
        Condenses trace entries into a single entry for each function call.
        """
        condensed = []
        active_functions = []  # Stack to track nested function calls
        function_entries = {}  # Store entries for each function

        for i, entry in enumerate(entries):
            function = entry["function"]
            
            if entry["type"] == "enter":
                # Initialize new function entry
                function_entries[function] = {
                    "depth": entry["depth"],
                    "function": function,
                    "timestamp": entry["timestamp"],
                    "inputs": None,
                    "output": None,
                    "evaluation_runs": [],
                    "span_type": entry.get("span_type", "span")
                }
                active_functions.append(function)
                
            elif entry["type"] == "exit" and function in active_functions:
                # Complete function entry
                current_entry = function_entries[function]
                current_entry["duration"] = entry["timestamp"] - current_entry["timestamp"]
                condensed.append(current_entry)
                active_functions.remove(function)
                # del function_entries[function]
                
            # The OR condition is to handle the LLM client case.
            # LLM client is a special case where we exit the span, so when we attach evaluations to it, 
            # we have to check if the previous entry is an LLM call.
            elif function in active_functions or entry["type"] == "evaluation" and entries[i-1]["function"] == entry["function"]:
                # Update existing function entry with additional data
                current_entry = function_entries[function]
                
                if entry["type"] == "input" and entry["inputs"]:
                    current_entry["inputs"] = entry["inputs"]
                    
                if entry["type"] == "output" and entry["output"]:
                    current_entry["output"] = entry["output"]
                    
                if entry["type"] == "evaluation" and entry["evaluation_runs"]:
                    current_entry["evaluation_runs"] = entry["evaluation_runs"]

        # Sort by timestamp
        condensed.sort(key=lambda x: x["timestamp"])
        
        return condensed

    def save(self, empty_save: bool = False, overwrite: bool = False) -> Tuple[str, dict]:
        """
        Save the current trace to the database.
        Returns a tuple of (trace_id, trace_data) where trace_data is the trace data that was saved.
        """
        # Calculate total elapsed time
        total_duration = self.get_duration()
        
        raw_entries = [entry.to_dict() for entry in self.entries]
        
        condensed_entries = self.condense_trace(raw_entries)

        # Calculate total token counts from LLM API calls
        total_prompt_tokens = 0
        total_completion_tokens = 0
        total_tokens = 0
        
        total_prompt_tokens_cost = 0.0
        total_completion_tokens_cost = 0.0
        total_cost = 0.0
        
        for entry in condensed_entries:
            if entry.get("span_type") == "llm" and isinstance(entry.get("output"), dict):
                output = entry["output"]
                usage = output.get("usage", {})
                model_name = entry.get("inputs", {}).get("model", "")
                prompt_tokens = 0
                completion_tokens = 0   
                
                # Handle OpenAI/Together format
                if "prompt_tokens" in usage:
                    prompt_tokens = usage.get("prompt_tokens", 0)
                    completion_tokens = usage.get("completion_tokens", 0)
                    total_prompt_tokens += prompt_tokens
                    total_completion_tokens += completion_tokens
                # Handle Anthropic format
                elif "input_tokens" in usage:
                    prompt_tokens = usage.get("input_tokens", 0)
                    completion_tokens = usage.get("output_tokens", 0)
                    total_prompt_tokens += prompt_tokens
                    total_completion_tokens += completion_tokens
                
                total_tokens += usage.get("total_tokens", 0)
                
                # Calculate costs if model name is available
                if model_name:
                    try:
                        prompt_cost, completion_cost = cost_per_token(
                            model=model_name, 
                            prompt_tokens=prompt_tokens, 
                            completion_tokens=completion_tokens
                        )
                        total_prompt_tokens_cost += prompt_cost
                        total_completion_tokens_cost += completion_cost
                        total_cost += prompt_cost + completion_cost
                        
                        # Add cost information directly to the usage dictionary in the condensed entry
                        if "usage" not in output:
                            output["usage"] = {}
                        output["usage"]["prompt_tokens_cost_usd"] = prompt_cost
                        output["usage"]["completion_tokens_cost_usd"] = completion_cost
                        output["usage"]["total_cost_usd"] = prompt_cost + completion_cost
                    except Exception as e:
                        # If cost calculation fails, continue without adding costs
                        print(f"Error calculating cost for model '{model_name}': {str(e)}")
                        pass

        # Create trace document
        trace_data = {
            "trace_id": self.trace_id,
            "name": self.name,
            "project_name": self.project_name,
            "created_at": datetime.utcfromtimestamp(self.start_time).isoformat(),
            "duration": total_duration,
            "token_counts": {
                "prompt_tokens": total_prompt_tokens,
                "completion_tokens": total_completion_tokens,
                "total_tokens": total_tokens,
                "prompt_tokens_cost_usd": total_prompt_tokens_cost,
                "completion_tokens_cost_usd": total_completion_tokens_cost,
                "total_cost_usd": total_cost
            },
            "entries": condensed_entries,
            "empty_save": empty_save,
            "overwrite": overwrite
        }
        # Execute asynchrous evaluation in the background
        if not empty_save:  # Only send to RabbitMQ if the trace is not empty
            # Send trace data to evaluation queue via API
            try:
                response = requests.post(
                    JUDGMENT_TRACES_ADD_TO_EVAL_QUEUE_API_URL,
                    json=trace_data,
                    headers={
                        "Content-Type": "application/json",
                        "Authorization": f"Bearer {self.tracer.api_key}",
                        "X-Organization-Id": self.tracer.organization_id
                    },
                    verify=True
                )
                
                if response.status_code != HTTPStatus.OK:
                    warnings.warn(f"Failed to add trace to evaluation queue: {response.text}")
            except Exception as e:
                warnings.warn(f"Error sending trace to evaluation queue: {str(e)}")
        
        self.trace_manager_client.save_trace(trace_data, empty_save)

        return self.trace_id, trace_data

    def delete(self):
        return self.trace_manager_client.delete_trace(self.trace_id)
    
class Tracer:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Tracer, cls).__new__(cls)
        return cls._instance

    def __init__(
        self, 
        api_key: str = os.getenv("JUDGMENT_API_KEY"), 
        project_name: str = "default_project",
        rules: Optional[List[Rule]] = None,  # Added rules parameter
        organization_id: str = os.getenv("JUDGMENT_ORG_ID"),
        enable_monitoring: bool = os.getenv("JUDGMENT_MONITORING", "true").lower() == "true",
        enable_evaluations: bool = os.getenv("JUDGMENT_EVALUATIONS", "true").lower() == "true"
        ):
        if not hasattr(self, 'initialized'):
            if not api_key:
                raise ValueError("Tracer must be configured with a Judgment API key")
            
            if not organization_id:
                raise ValueError("Tracer must be configured with an Organization ID")
            self.api_key: str = api_key
            self.project_name: str = project_name
            self.client: JudgmentClient = JudgmentClient(judgment_api_key=api_key)
            self.organization_id: str = organization_id
            self.depth: int = 0
            self._current_trace: Optional[str] = None
            self.rules: List[Rule] = rules or []  # Store rules at tracer level
            self.initialized: bool = True
            self.enable_monitoring: bool = enable_monitoring
            self.enable_evaluations: bool = enable_evaluations
        elif hasattr(self, 'project_name') and self.project_name != project_name:
            warnings.warn(
                f"Attempting to initialize Tracer with project_name='{project_name}' but it was already initialized with "
                f"project_name='{self.project_name}'. Due to the singleton pattern, the original project_name will be used. "
                "To use a different project name, ensure the first Tracer initialization uses the desired project name.",
                RuntimeWarning
            )
        
    @contextmanager
    def trace(
        self, 
        name: str, 
        project_name: str = None, 
        overwrite: bool = False,
        rules: Optional[List[Rule]] = None  # Added rules parameter
    ) -> Generator[TraceClient, None, None]:
        """Start a new trace context using a context manager"""
        trace_id = str(uuid.uuid4())
        project = project_name if project_name is not None else self.project_name

        trace = TraceClient(
            self, 
            trace_id, 
            name, 
            project_name=project, 
            overwrite=overwrite,
            rules=self.rules,  # Pass combined rules to the trace client
            enable_monitoring=self.enable_monitoring,
            enable_evaluations=self.enable_evaluations
        )
        prev_trace = self._current_trace
        self._current_trace = trace
        
        # Automatically create top-level span
        with trace.span(name or "unnamed_trace") as span:
            try:
                # Save the trace to the database to handle Evaluations' trace_id referential integrity
                trace.save(empty_save=True, overwrite=overwrite)
                yield trace
            finally:
                self._current_trace = prev_trace
                
    def get_current_trace(self) -> Optional[TraceClient]:
        """
        Get the current trace context
        """
        return self._current_trace    

    def observe(self, func=None, *, name=None, span_type: SpanType = "span", project_name: str = None, overwrite: bool = False):
        """
        Decorator to trace function execution with detailed entry/exit information.
        
        Args:
            func: The function to decorate
            name: Optional custom name for the span (defaults to function name)
            span_type: Type of span (default "span")
            project_name: Optional project name override
            overwrite: Whether to overwrite existing traces
        """
        # If monitoring is disabled, return the function as is
        if not self.enable_monitoring:
            return func if func else lambda f: f
        
        if func is None:
            return lambda f: self.observe(f, name=name, span_type=span_type, project_name=project_name, overwrite=overwrite)
        
        # Use provided name or fall back to function name
        span_name = name or func.__name__
        
        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                # If there's already a trace, use it. Otherwise create a new one
                if self._current_trace:
                    trace = self._current_trace
                else:
                    trace_id = str(uuid.uuid4())
                    trace_name = func.__name__
                    project = project_name if project_name is not None else self.project_name
                    trace = TraceClient(self, trace_id, trace_name, project_name=project, overwrite=overwrite, rules=self.rules, enable_monitoring=self.enable_monitoring, enable_evaluations=self.enable_evaluations)
                    self._current_trace = trace
                    # Only save empty trace for the root call
                    trace.save(empty_save=True, overwrite=overwrite)

                try:
                    with trace.span(span_name, span_type=span_type) as span:
                        # Record inputs
                        span.record_input({
                            'args': str(args),
                            'kwargs': kwargs
                        })
                        
                        # Execute function
                        result = await func(*args, **kwargs)
                        
                        # Record output
                        span.record_output(result)
                        
                        return result
                finally:
                    # Only save and cleanup if this is the root observe call
                    if self.depth == 0:
                        trace.save(empty_save=False, overwrite=overwrite)
                        self._current_trace = None
                    
            return async_wrapper
        else:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # If there's already a trace, use it. Otherwise create a new one
                if self._current_trace:
                    trace = self._current_trace
                else:
                    trace_id = str(uuid.uuid4())
                    trace_name = func.__name__
                    project = project_name if project_name is not None else self.project_name
                    trace = TraceClient(self, trace_id, trace_name, project_name=project, overwrite=overwrite, rules=self.rules, enable_monitoring=self.enable_monitoring)
                    self._current_trace = trace
                    # Only save empty trace for the root call
                    trace.save(empty_save=True, overwrite=overwrite)
                
                try:
                    with trace.span(span_name, span_type=span_type) as span:
                        # Record inputs
                        span.record_input({
                            'args': str(args),
                            'kwargs': kwargs
                        })
                        
                        # Execute function
                        result = func(*args, **kwargs)
                        
                        # Record output
                        span.record_output(result)
                        
                        return result
                finally:
                    # Only save and cleanup if this is the root observe call
                    if self.depth == 0:
                        trace.save(empty_save=False, overwrite=overwrite)
                        self._current_trace = None
                    
            return wrapper
        
    def score(self, func=None, scorers: List[Union[APIJudgmentScorer, JudgevalScorer]] = None, model: str = None, log_results: bool = True, *, name: str = None, span_type: SpanType = "span"):
        """
        Decorator to trace function execution with detailed entry/exit information.
        """
        if func is None:
            return lambda f: self.observe(f, name=name, span_type=span_type)
        
        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                if self._current_trace:
                    self._current_trace.async_evaluate(scorers=[scorers], input=args, actual_output=kwargs, model=model, log_results=log_results)
            return async_wrapper
        else:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if self._current_trace:
                    self._current_trace.async_evaluate(scorers=[scorers], input=args, actual_output=kwargs, model="gpt-4o-mini", log_results=True)
            return wrapper
        
    def async_evaluate(self, *args, **kwargs):
        if not self.enable_evaluations:
            return

        if self._current_trace:
            self._current_trace.async_evaluate(*args, **kwargs)
        else:
            warnings.warn("No trace found, skipping evaluation")


def wrap(client: Any) -> Any:
    """
    Wraps an API client to add tracing capabilities.
    Supports OpenAI, Together, and Anthropic clients.
    """
    # Get the appropriate configuration for this client type
    span_name, original_create = _get_client_config(client)
    
    def traced_create(*args, **kwargs):
        # Get the current tracer instance (might be created after client was wrapped)
        tracer = Tracer._instance
        
        # Skip tracing if no tracer exists or no active trace
        if not tracer or not tracer._current_trace:
            return original_create(*args, **kwargs)

        with tracer._current_trace.span(span_name, span_type="llm") as span:
            # Format and record the input parameters
            input_data = _format_input_data(client, **kwargs)
            span.record_input(input_data)
            
            # Make the actual API call
            response = original_create(*args, **kwargs)
            
            # Format and record the output
            output_data = _format_output_data(client, response)
            span.record_output(output_data)
            
            return response
            
    # Replace the original method with our traced version
    if isinstance(client, (OpenAI, Together)):
        client.chat.completions.create = traced_create
    elif isinstance(client, Anthropic):
        client.messages.create = traced_create
        
    return client

# Helper functions for client-specific operations

def _get_client_config(client: ApiClient) -> tuple[str, callable]:
    """Returns configuration tuple for the given API client.
    
    Args:
        client: An instance of OpenAI, Together, or Anthropic client
        
    Returns:
        tuple: (span_name, create_method)
            - span_name: String identifier for tracing
            - create_method: Reference to the client's creation method
            
    Raises:
        ValueError: If client type is not supported
    """
    if isinstance(client, OpenAI):
        return "OPENAI_API_CALL", client.chat.completions.create
    elif isinstance(client, Together):
        return "TOGETHER_API_CALL", client.chat.completions.create
    elif isinstance(client, Anthropic):
        return "ANTHROPIC_API_CALL", client.messages.create
    raise ValueError(f"Unsupported client type: {type(client)}")

def _format_input_data(client: ApiClient, **kwargs) -> dict:
    """Format input parameters based on client type.
    
    Extracts relevant parameters from kwargs based on the client type
    to ensure consistent tracing across different APIs.
    """
    if isinstance(client, (OpenAI, Together)):
        return {
            "model": kwargs.get("model"),
            "messages": kwargs.get("messages"),
        }
    # Anthropic requires additional max_tokens parameter
    return {
        "model": kwargs.get("model"),
        "messages": kwargs.get("messages"),
        "max_tokens": kwargs.get("max_tokens")
    }

def _format_output_data(client: ApiClient, response: Any) -> dict:
    """Format API response data based on client type.
    
    Normalizes different response formats into a consistent structure
    for tracing purposes.
    
    Returns:
        dict containing:
            - content: The generated text
            - usage: Token usage statistics
    """
    if isinstance(client, (OpenAI, Together)):
        return {
            "content": response.choices[0].message.content,
            "usage": {
                "prompt_tokens": response.usage.prompt_tokens,
                "completion_tokens": response.usage.completion_tokens,
                "total_tokens": response.usage.total_tokens
            }
        }
    # Anthropic has a different response structure
    return {
        "content": response.content[0].text,
        "usage": {
            "input_tokens": response.usage.input_tokens,
            "output_tokens": response.usage.output_tokens,
            "total_tokens": response.usage.input_tokens + response.usage.output_tokens
        }
    }


"""
Util functions for Scorer objects
"""

import asyncio
import nest_asyncio
import inspect
import json
import sys
import re
from contextlib import contextmanager
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.console import Console
from typing import List, Optional

from judgeval.scorers import JudgevalScorer
from judgeval.data import Example, ExampleParams
from judgeval.scorers.exceptions import MissingExampleParamsError


def clone_scorers(scorers: List[JudgevalScorer]) -> List[JudgevalScorer]:
    """
    Creates duplicates of the scorers passed as argument.
    """
    cloned_scorers = []
    for s in scorers:
        scorer_class = type(s)
        args = vars(s)

        signature = inspect.signature(scorer_class.__init__)
        valid_params = signature.parameters.keys()
        valid_args = {key: args[key] for key in valid_params if key in args}

        cloned_scorer = scorer_class(**valid_args)
        # kinda hacky, but in case the class inheriting from JudgevalScorer doesn't have `model` in its __init__,
        # we need to explicitly include it here so that we can add the judge model to the cloned scorer
        cloned_scorer._add_model(model=args.get("model"))
        cloned_scorers.append(cloned_scorer)
    return cloned_scorers


def scorer_console_msg(
    scorer: JudgevalScorer,
    async_mode: Optional[bool] = None,
):
    """
    Renders a message to be displayed to console when a scorer is being executed.
    """
    if async_mode is None:
        run_async = scorer.async_mode
    else:
        run_async = async_mode

    return f"🔨 Executing Judgment's [rgb(106,0,255)]{scorer.__name__} Scorer[/rgb(106,0,255)]! \
        [rgb(55,65,81)](using {scorer.evaluation_model}, async_mode={run_async})...[/rgb(55,65,81)]"


@contextmanager
def scorer_progress_meter(
    scorer: JudgevalScorer,
    async_mode: Optional[bool] = None,
    display_meter: bool = True,
    total: int = 100,
    transient: bool = True,
):
    """
    Context manager to display a progress indicator (spinner) while a scorer is being run.
    """
    console = Console(file=sys.stderr)
    if display_meter:
        with Progress(
            SpinnerColumn(style="rgb(106,0,255)"),
            TextColumn("[progress.description]{task.description}"),
            console=console,
            transient=transient,
        ) as progress:
            progress.add_task(
                description=scorer_console_msg(scorer, async_mode),
                total=total,
            )
            yield
    else:
        yield


def parse_response_json(llm_response: str, scorer: Optional[JudgevalScorer] = None) -> dict:
    """
    Extracts JSON output from an LLM response and returns it as a dictionary.

    If the JSON is invalid, the error is forwarded to the `scorer`, if provided.

    Args:
        llm_response (str): The response from an LLM.
        scorer (JudgevalScorer, optional): The scorer object to forward errors to (if any).
    """
    start = llm_response.find("{")  # opening bracket
    end = llm_response.rfind("}") + 1  # closing bracket

    if end == 0 and start != -1:  # add the closing bracket if it's missing
        llm_response = llm_response + "}"
        end = len(llm_response)

    json_str = llm_response[start:end] if start != -1 and end != 0 else ""  # extract the JSON string
    json_str = re.sub(r",\s*([\]}])", r"\1", json_str)  # Remove trailing comma if present

    try:
        return json.loads(json_str)
    except json.JSONDecodeError:
        error_str = "Evaluation LLM outputted an invalid JSON. Please use a stronger evaluation model."
        if scorer is not None:
            scorer.error = error_str
        raise ValueError(error_str)
    except Exception as e:
        raise Exception(f"An unexpected error occurred: {str(e)}")


def print_verbose_logs(metric: str, logs: str):
    print("*" * 50)
    print(f"{metric} Verbose Logs")
    print("*" * 50)
    print("")
    print(logs)
    print("")
    print("=" * 70)


def create_verbose_logs(metric: JudgevalScorer, steps: List[str]) -> str:
    """
    Creates verbose logs for a scorer object.

    Args:
        metric (JudgevalScorer): The scorer object.
        steps (List[str]): The steps to be included in the verbose logs.
    
    Returns:
        str: The verbose logs (Concatenated steps).
    """

    verbose_logs = ""
    for i in range(len(steps) - 1):
        verbose_logs += steps[i]
        if i < len(steps) - 2:  # don't add new line for penultimate step
            verbose_logs += " \n \n"
    if metric.verbose_mode:
        print_verbose_logs(metric.__name__, verbose_logs + f"\n \n{steps[-1]}")
    return verbose_logs


def get_or_create_event_loop() -> asyncio.AbstractEventLoop:
    """
    Get or create an asyncio event loop.

    This function attempts to retrieve the current event loop using `asyncio.get_event_loop()`.
    If the event loop is already running, it applies the `nest_asyncio` patch to allow nested
    asynchronous execution. If the event loop is closed or not found, it creates a new event loop
    and sets it as the current event loop.

    Returns:
        asyncio.AbstractEventLoop: The current or newly created event loop.
    
    Raises:
        RuntimeError: If the event loop is closed.
    """
    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            print(
                "Event loop is already running. Applying nest_asyncio patch to allow async execution..."
            )
            nest_asyncio.apply()

        if loop.is_closed():
            raise RuntimeError
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def check_example_params(
    example: Example,
    example_params: List[ExampleParams],
    scorer: JudgevalScorer,
):
    if isinstance(example, Example) is False:
        error_str = f"in check_example_params(): Expected example to be of type 'Example', but got {type(example)}"
        scorer.error = error_str
        raise MissingExampleParamsError(error_str)

    missing_params = []
    for param in example_params:
        if getattr(example, param.value) is None:
            missing_params.append(f"'{param.value}'")

    if missing_params:
        if len(missing_params) == 1:
            missing_params_str = missing_params[0]
        elif len(missing_params) == 2:
            missing_params_str = " and ".join(missing_params)
        else:
            missing_params_str = (
                ", ".join(missing_params[:-1]) + ", and " + missing_params[-1]
            )

        error_str = f"{missing_params_str} fields in example cannot be None for the '{scorer.__name__}' scorer"
        scorer.error = error_str
        raise MissingExampleParamsError(error_str)





"""
Judgment Scorer class.

Scores `Example`s using ready-made Judgment evaluators.
"""

from pydantic import BaseModel, field_validator
from judgeval.common.logger import debug, info, warning, error

from judgeval.constants import APIScorer, UNBOUNDED_SCORERS


class APIJudgmentScorer(BaseModel):
    """
    Class for ready-made, "out-of-the-box" scorer that uses Judgment evaluators to score `Example`s.

    Args:
        score_type (APIScorer): The Judgment metric to use for scoring `Example`s
        threshold (float): A value between 0 and 1 that determines the scoring threshold
    """
    score_type: APIScorer
    threshold: float

    @field_validator('threshold')
    def validate_threshold(cls, v, info):
        """
        Validates that the threshold is between 0 and 1 inclusive.
        """
        score_type = info.data.get('score_type')
        if score_type in UNBOUNDED_SCORERS:
            if v < 0:
                error(f"Threshold for {score_type} must be greater than 0, got: {v}")
                raise ValueError(f"Threshold for {score_type} must be greater than 0, got: {v}")
        else:
            if not 0 <= v <= 1:
                error(f"Threshold for {score_type} must be between 0 and 1, got: {v}")
                raise ValueError(f"Threshold for {score_type} must be between 0 and 1, got: {v}")
        return v

    @field_validator('score_type')
    def convert_to_enum_value(cls, v):
        """
        Validates that the `score_type` is a valid `JudgmentMetric` enum value.
        Converts string values to `JudgmentMetric` enum values.
        """
        debug(f"Attempting to convert score_type value: {v}")
        if isinstance(v, APIScorer):
            info(f"Using existing JudgmentMetric: {v.value}")
            return v.value
        elif isinstance(v, str):
            debug(f"Converting string value to JudgmentMetric enum: {v}")
            return APIScorer[v.upper()].value
        error(f"Invalid score_type value: {v}")
        raise ValueError(f"Invalid value for score_type: {v}")
    
    def __str__(self):
        return f"JudgmentScorer(score_type={self.score_type.value}, threshold={self.threshold})"
    


from typing import Any, Dict, List, Optional, Sequence
from uuid import UUID
import time
import uuid
from contextvars import ContextVar
from judgeval.common.tracer import TraceClient, TraceEntry, Tracer, SpanType

from langchain_core.language_models import BaseChatModel
from langchain_huggingface import ChatHuggingFace
from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_core.callbacks import CallbackManager, BaseCallbackHandler
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.outputs import LLMResult
from langchain_core.tracers.context import register_configure_hook
from langchain_core.messages.ai import AIMessage
from langchain_core.messages.tool import ToolMessage
from langchain_core.messages.base import BaseMessage
from langchain_core.documents import Document

class JudgevalCallbackHandler(BaseCallbackHandler):
    def __init__(self, tracer: Tracer):
        self.tracer = tracer
        self.trace_client = tracer.get_current_trace() if tracer.get_current_trace() else None
        self.previous_spans = [] # stack of previous spans
        self.finished = False

        # Attributes for users to access
        self.previous_node = None
        self.executed_node_tools = []
        self.executed_nodes = []
        self.executed_tools = []

    def start_span(self, name: str, span_type: SpanType = "span"):
        start_time = time.time()
        
        # Record span entry
        self.trace_client.add_entry(TraceEntry(
            type="enter",
            function=name,
            depth=self.trace_client.tracer.depth,
            message=name,
            timestamp=start_time,
            span_type=span_type
        ))

        self.trace_client.tracer.depth += 1
        self.previous_spans.append(self.trace_client._current_span)
        self.trace_client._current_span = name
        self._start_time = start_time
    
    def end_span(self, name: str, span_type: SpanType = "span"):
        self.trace_client.tracer.depth -= 1
        duration = time.time() - self._start_time
        
        # Record span exit
        self.trace_client.add_entry(TraceEntry(
            type="exit",
            function=name,
            depth=self.trace_client.tracer.depth,
            message=f"{name}",
            timestamp=time.time(),
            duration=duration,
            span_type=span_type
        ))
        self.trace_client._current_span = self.previous_spans.pop()

        if self.trace_client.tracer.depth == 0:
            # Save the trace if we are the root, this is when users dont use any @observe decorators
            self.trace_client.save(empty_save=False, overwrite=True)
            self.trace_client._current_trace = None
                 
    def on_retriever_start(
        self,
        serialized: Optional[dict[str, Any]],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        name = "RETRIEVER_CALL"
        if serialized and "name" in serialized:
            name = f"RETRIEVER_{serialized['name'].upper()}"
        
        self.start_span(name, span_type="retriever")
        self.trace_client.record_input({
            'query': query,
            'tags': tags,
            'metadata': metadata,
            'kwargs': kwargs
        })

    def on_retriever_end(
        self, 
        documents: Sequence[Document], 
        *, 
        run_id: UUID, 
        parent_run_id: Optional[UUID] = None, 
        **kwargs: Any
    ) -> Any:
        # Process the retrieved documents into a format suitable for logging
        doc_summary = []
        for i, doc in enumerate(documents):
            # Extract key information from each document
            doc_data = {
                "index": i,
                "page_content": doc.page_content[:100] + "..." if len(doc.page_content) > 100 else doc.page_content,
                "metadata": doc.metadata
            }
            doc_summary.append(doc_data)
        
        # Record the document data
        self.trace_client.record_output({
            "document_count": len(documents),
            "documents": doc_summary
        })
        
        # End the retriever span
        self.end_span(self.trace_client._current_span, span_type="retriever")

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any
    ) -> None:
        # If the user doesnt use any @observe decorators, the first action in LangGraph workflows seems tohave this attribute, so we intialize our trace client here
        if kwargs.get('name') == 'LangGraph':
            if not self.trace_client:
                trace_id = str(uuid.uuid4())
                project = self.tracer.project_name
                trace = TraceClient(self.tracer, trace_id, trace_id, project_name=project, overwrite=False, rules=self.tracer.rules, enable_monitoring=self.tracer.enable_monitoring, enable_evaluations=self.tracer.enable_evaluations)
                self.trace_client = trace
                self.tracer._current_trace = trace # set the trace in the original tracer object
                # Only save empty trace for the root call
                self.trace_client.save(empty_save=True, overwrite=False)
            
            self.start_span("LangGraph", span_type="Main Function")
            
        metadata = kwargs.get("metadata", {})
        if node := metadata.get("langgraph_node"):
            if node != self.previous_node:
                # Track node execution
                self.trace_client.visited_nodes.append(node)
                self.trace_client.executed_node_tools.append(node)
                self.trace_client.record_input({
                    'args': inputs,
                    'kwargs': kwargs
                })
            self.previous_node = node
    
    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Any:   
        if outputs == "__end__":
            self.finished = True
        if tags is not None and any("graph:step" in tag for tag in tags):
            self.trace_client.record_output(outputs)
            self.end_span(self.trace_client._current_span, span_type="node")

            if self.finished:
                self.end_span(self.trace_client._current_span, span_type="Main Function")
    
    def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print(f"Chain error: {error}")
        self.trace_client.record_output(error)
        self.end_span(self.trace_client._current_span, span_type="node")

    def on_tool_start(
        self,
        serialized: Optional[dict[str, Any]],
        input_str: str,
        run_id: Optional[UUID] = None,
        parent_run_id: Optional[UUID] = None,
        inputs: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ):
        name = serialized["name"]
        self.start_span(name, span_type="tool")
        if name:
            # Track tool execution
            self.trace_client.executed_tools.append(name)
            node_tool = f"{self.previous_node}:{name}" if self.previous_node else name
            self.trace_client.executed_node_tools.append(node_tool)
        self.trace_client.record_input({
            'args': input_str,
            'kwargs': kwargs
        })

    def on_tool_end(self, output: Any, *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any) -> Any:
        self.trace_client.record_output(output)
        self.end_span(self.trace_client._current_span, span_type="tool")

    def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print(f"Tool error: {error}")
        self.trace_client.record_output(error)
        self.end_span(self.trace_client._current_span, span_type="tool")

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print(f"Agent action: {action}")

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print(f"Agent finish: {finish}")

    def on_llm_start(
        self,
        serialized: Optional[dict[str, Any]],
        prompts: list[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:    
        name = "LLM call"
        self.start_span(name, span_type="llm")
        self.trace_client.record_input({
            'args': prompts,
            'kwargs': kwargs
        })

    def on_llm_end(self, response: LLMResult, *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any):
        self.trace_client.record_output(response.generations[0][0].text)
        self.end_span(self.trace_client._current_span, span_type="llm")

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print(f"LLM error: {error}")
        self.trace_client.record_output(error)
        self.end_span(self.trace_client._current_span, span_type="llm")

    def on_chat_model_start(
        self,
        serialized: Optional[dict[str, Any]],
        messages: list[list[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:

        if "openai" in serialized["id"]:
            name = f"OPENAI_API_CALL"
        elif "anthropic" in serialized["id"]:
            name = "ANTHROPIC_API_CALL"
        elif "together" in serialized["id"]:
            name = "TOGETHER_API_CALL"
        else:
            name = "LLM call"

        self.start_span(name, span_type="llm")
        self.trace_client.record_input({
            'args': str(messages),
            'kwargs': kwargs
        })

judgeval_callback_handler_var: ContextVar[Optional[JudgevalCallbackHandler]] = ContextVar(
    "judgeval_callback_handler", default=None
)

def set_global_handler(handler: JudgevalCallbackHandler):
    if not handler.tracer.enable_monitoring:
        return
    judgeval_callback_handler_var.set(handler)

def clear_global_handler():
    judgeval_callback_handler_var.set(None)

register_configure_hook(
    context_var=judgeval_callback_handler_var,
    inheritable=True,
)


import logging
from logging.handlers import RotatingFileHandler
import sys
from pathlib import Path
from contextlib import contextmanager

# Global variables
logger = None
class LoggingState:
    enabled = False
    path = None

LOGGING_STATE = LoggingState()

# Add these as module-level variables
current_example_id = None
current_timestamp = None


@contextmanager
def enable_logging(name: str = "judgeval", path: str = "./logs", max_bytes: int = 1024 * 1024, backup_count: int = 5):
    """
    Context manager to temporarily enable logging for a specific block of code.
    """
    global logger
    LOGGING_STATE.enabled = True
    LOGGING_STATE.path = path
    # Initialize logger if not already initialized
    if logger is None:
        logger = _initialize_logger(name=name, path=path, max_bytes=max_bytes, backup_count=backup_count)
    try:
        logger.info("Logging enabled")
        yield
    finally:
        logger.info("Logging disabled")
        LOGGING_STATE.enabled = False
        LOGGING_STATE.path = None

def _initialize_logger(
    name: str = "judgeval",
    max_bytes: int = 1024 * 1024,  # 1MB
    backup_count: int = 5,
    path: str = "./logs"  # Added path parameter with default
) -> logging.Logger:
    """
    Initialize the global logger instance if it doesn't exist.
    Returns the global logger instance.
    """
    global logger
    
    log_dir = Path(path)
    log_dir.mkdir(exist_ok=True, parents=True)
    log_file = log_dir / f"{name}.log"
    if log_file.exists():
        log_file.unlink()  # Delete existing log file
    
    if logger is not None:
        return logger
        
    # Create logs directory if it doesn't exist
    log_dir = Path(path)
    log_dir.mkdir(exist_ok=True)
    
    # Create formatter
    formatter = logging.Formatter(
        fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    # Create a custom formatter that includes example info when available
    class ExampleFormatter(logging.Formatter):
        def format(self, record):
            if current_example_id is not None and current_timestamp is not None:
                record.example_id = current_example_id
                record.timestamp = current_timestamp
                return logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - [Example_%(example_id)s][%(timestamp)s] %(message)s', 
                                      datefmt='%Y-%m-%d %H:%M:%S').format(record)
            return logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S').format(record)
    
    # Use the custom formatter
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(ExampleFormatter())
    console_handler.setLevel(logging.DEBUG)
    
    log_filename = f"{name}.log"
    file_handler = RotatingFileHandler(
        log_dir / log_filename,
        maxBytes=max_bytes,
        backupCount=backup_count,
        mode='a'
    )
    file_handler.setFormatter(ExampleFormatter())
    file_handler.setLevel(logging.DEBUG)

    # Get logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    
    # Prevent adding handlers multiple times
    if not logger.handlers:
        logger.addHandler(console_handler)
        logger.addHandler(file_handler)
    
    return logger

# Initialize the global logger when module is imported
# logger = _initialize_logger()

def log_if_enabled(func):
    """Decorator to check if logging is enabled before executing logging statements"""
    def wrapper(*args, **kwargs):
        if LOGGING_STATE.enabled:
            return func(*args, **kwargs)
    return wrapper

@log_if_enabled
def debug(msg: str, example_idx: int = None):
    """Log debug message if logging is enabled"""
    logger.debug(msg)

@log_if_enabled
def info(msg: str, example_idx: int = None):
    """Log info message if logging is enabled"""
    logger.info(msg)

@log_if_enabled
def warning(msg: str, example_idx: int = None):
    """Log warning message if logging is enabled"""
    logger.warning(msg)

@log_if_enabled
def error(msg: str, example_idx: int = None):
    """Log error message if logging is enabled"""
    logger.error(msg)

def create_example_handler(
    timestamp: str, 
    example_idx: int,
    path: str = "./logs"  # Added path parameter with default
) -> RotatingFileHandler:
    """Creates a file handler for a specific example"""
    debug(f"Creating example handler for timestamp={timestamp}, example_idx={example_idx}")
    log_dir = Path(path) / "examples"
    log_dir.mkdir(exist_ok=True, parents=True)
    
    formatter = logging.Formatter(
        fmt='%(asctime)s - %(name)s - %(levelname)s - [Example_%(example_id)s][%(timestamp)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    
    # Create a unique file for each example
    file_handler = RotatingFileHandler(
        log_dir / f"{timestamp}_example_{example_idx}.log",
        maxBytes=1024 * 1024,  # 1MB
        backupCount=5,
        mode='a'
    )
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)
    info(f"Created example handler for example {example_idx}")
    return file_handler

@contextmanager
def example_logging_context(timestamp: str, example_idx: int):
    """Context manager for example-specific logging"""
    if not LOGGING_STATE.enabled:
        yield
        return
        
    global current_example_id, current_timestamp
    
    debug(f"Entering example logging context for example {example_idx}")
    current_example_id = example_idx
    current_timestamp = timestamp
    
    handler = create_example_handler(timestamp, example_idx, path=LOGGING_STATE.path)
    if handler:
        logger.addHandler(handler)
    try:
        yield
    finally:
        current_example_id = None
        current_timestamp = None
        if handler:
            logger.removeHandler(handler)
            handler.close()
            debug(f"Closed example handler for example {example_idx}")


from uuid import uuid4
import openai
import os
import asyncio
from tavily import TavilyClient
from dotenv import load_dotenv
import chromadb
from chromadb.utils import embedding_functions

from judgeval.tracer import Tracer, wrap
from judgeval.scorers import AnswerRelevancyScorer, FaithfulnessScorer

destinations_data = [
    {
        "destination": "Paris, France",
        "information": """
Paris is the capital city of France and a global center for art, fashion, and culture.
Key Information:
- Best visited during spring (March-May) or fall (September-November)
- Famous landmarks: Eiffel Tower, Louvre Museum, Notre-Dame Cathedral, Arc de Triomphe
- Known for: French cuisine, café culture, fashion, art galleries
- Local transportation: Metro system is extensive and efficient
- Popular neighborhoods: Le Marais, Montmartre, Latin Quarter
- Cultural tips: Basic French phrases are appreciated; many restaurants close between lunch and dinner
- Must-try experiences: Seine River cruise, visiting local bakeries, Luxembourg Gardens
"""
    },
    {
        "destination": "Tokyo, Japan",
        "information": """
Tokyo is Japan's bustling capital, blending ultramodern and traditional elements.
Key Information:
- Best visited during spring (cherry blossoms) or fall (autumn colors)
- Famous areas: Shibuya, Shinjuku, Harajuku, Akihabara
- Known for: Technology, anime culture, sushi, efficient public transport
- Local transportation: Extensive train and subway network
- Cultural tips: Bow when greeting, remove shoes indoors, no tipping
- Must-try experiences: Robot Restaurant, teamLab Borderless, Tsukiji Outer Market
- Popular day trips: Mount Fuji, Kamakura, Nikko
"""
    },
    {
        "destination": "New York City, USA",
        "information": """
New York City is a global metropolis known for its diversity, culture, and iconic skyline.
Key Information:
- Best visited during spring (April-June) or fall (September-November)
- Famous landmarks: Statue of Liberty, Times Square, Central Park, Empire State Building
- Known for: Broadway shows, diverse cuisine, shopping, museums
- Local transportation: Extensive subway system, yellow cabs, ride-sharing
- Popular areas: Manhattan, Brooklyn, Queens
- Cultural tips: Fast-paced environment, tipping expected (15-20%)
- Must-try experiences: Broadway show, High Line walk, food tours
"""
    },
    {
        "destination": "Barcelona, Spain",
        "information": """
Barcelona is a vibrant city known for its art, architecture, and Mediterranean culture.
Key Information:
- Best visited during spring and fall for mild weather
- Famous landmarks: Sagrada Familia, Park Güell, Casa Batlló
- Known for: Gaudi architecture, tapas, beach culture, FC Barcelona
- Local transportation: Metro, buses, and walkable city center
- Popular areas: Gothic Quarter, Eixample, La Barceloneta
- Cultural tips: Late dinner times (after 8 PM), siesta tradition
- Must-try experiences: La Rambla walk, tapas crawl, local markets
"""
    },
    {
        "destination": "Bangkok, Thailand",
        "information": """
Bangkok is Thailand's capital city, famous for its temples, street food, and vibrant culture.
Key Information:
- Best visited during November to February (cool and dry season)
- Famous sites: Grand Palace, Wat Phra Kaew, Wat Arun
- Known for: Street food, temples, markets, nightlife
- Local transportation: BTS Skytrain, MRT, tuk-tuks, river boats
- Popular areas: Sukhumvit, Old City, Chinatown
- Cultural tips: Dress modestly at temples, respect royal family
- Must-try experiences: Street food tours, river cruises, floating markets
"""
    }
]

client = wrap(openai.Client(api_key=os.getenv("OPENAI_API_KEY")))
judgment = Tracer(api_key=os.getenv("JUDGMENT_API_KEY"), project_name="travel_agent_demo")

def populate_vector_db(collection, destinations_data):
    """
    Populate the vector DB with travel information.
    destinations_data should be a list of dictionaries with 'destination' and 'information' keys
    """
    for data in destinations_data:
        collection.add(
            documents=[data['information']],
            metadatas=[{"destination": data['destination']}],
            ids=[f"destination_{data['destination'].lower().replace(' ', '_')}"]
        )

@judgment.observe(span_type="search_tool")
def search_tavily(query):
    """Fetch travel data using Tavily API."""
    API_KEY = os.getenv("TAVILY_API_KEY")
    client = TavilyClient(api_key=API_KEY)
    results = client.search(query, num_results=3)
    return results

@judgment.observe(span_type="tool")
async def get_attractions(destination):
    """Search for top attractions in the destination."""
    prompt = f"Best tourist attractions in {destination}"
    attractions_search = search_tavily(prompt)
    return attractions_search

@judgment.observe(span_type="tool")
async def get_hotels(destination):
    """Search for hotels in the destination."""
    prompt = f"Best hotels in {destination}"
    hotels_search = search_tavily(prompt)
    return hotels_search

@judgment.observe(span_type="tool")
async def get_flights(destination):
    """Search for flights to the destination."""
    prompt = f"Flights to {destination} from major cities"
    flights_search = search_tavily(prompt)
    judgment.async_evaluate(
        scorers=[AnswerRelevancyScorer(threshold=0.5)],
        input=prompt,
        actual_output=str(flights_search["results"]),
        model="gpt-4o",
    )
    return flights_search

@judgment.observe(span_type="tool")
async def get_weather(destination, start_date, end_date):
    """Search for weather information."""
    prompt = f"Weather forecast for {destination} from {start_date} to {end_date}"
    weather_search = search_tavily(prompt)
    judgment.async_evaluate(
        scorers=[AnswerRelevancyScorer(threshold=0.5)],
        input=prompt,
        actual_output=str(weather_search["results"]),
        model="gpt-4o",
    )
    return weather_search

def initialize_vector_db():
    """Initialize ChromaDB with OpenAI embeddings."""
    client = chromadb.Client()
    embedding_fn = embedding_functions.OpenAIEmbeddingFunction(
        api_key=os.getenv("OPENAI_API_KEY"),
        model_name="text-embedding-3-small"
    )
    res = client.get_or_create_collection(
        "travel_information",
        embedding_function=embedding_fn
    )
    populate_vector_db(res, destinations_data)
    return res

@judgment.observe(span_type="retriever")
def query_vector_db(collection, destination, k=3):
    """Query the vector database for existing travel information."""
    try:
        results = collection.query(
            query_texts=[destination],
            n_results=k
        )
        return results['documents'][0] if results['documents'] else []
    except Exception:
        return []

@judgment.observe(span_type="Research")
async def research_destination(destination, start_date, end_date):
    """Gather all necessary travel information for a destination."""
    # First, check the vector database
    collection = initialize_vector_db()
    existing_info = query_vector_db(collection, destination)
    
    # Get real-time information from Tavily
    tavily_data = {
        "attractions": await get_attractions(destination),
        "hotels": await get_hotels(destination),
        "flights": await get_flights(destination),
        "weather": await get_weather(destination, start_date, end_date)
    }
    
    return {
        "vector_db_results": existing_info,
        **tavily_data
    }

@judgment.observe(span_type="function")
async def create_travel_plan(destination, start_date, end_date, research_data):
    """Generate a travel itinerary using the researched data."""
    vector_db_context = "\n".join(research_data['vector_db_results']) if research_data['vector_db_results'] else "No pre-stored information available."
    
    prompt = f"""
    Create a structured travel itinerary for a trip to {destination} from {start_date} to {end_date}.
    
    Pre-stored destination information:
    {vector_db_context}
    
    Current travel data:
    - Attractions: {research_data['attractions']}
    - Hotels: {research_data['hotels']}
    - Flights: {research_data['flights']}
    - Weather: {research_data['weather']}
    """
    
    response = client.chat.completions.create(
        model="gpt-4o",
        messages=[
            {"role": "system", "content": "You are an expert travel planner. Combine both historical and current information to create the best possible itinerary."},
            {"role": "user", "content": prompt}
        ]
    ).choices[0].message.content

    judgment.async_evaluate(
        scorers=[FaithfulnessScorer(threshold=0.5)],
        input=prompt,
        actual_output=str(response),
        retrieval_context=[str(vector_db_context), str(research_data)],
        model="gpt-4o",
    )
    
    return response

@judgment.observe(span_type="function")
async def generate_itinerary(destination, start_date, end_date):
    """Main function to generate a travel itinerary."""
    research_data = await research_destination(destination, start_date, end_date)
    res = await create_travel_plan(destination, start_date, end_date, research_data)
    return res


if __name__ == "__main__":
    load_dotenv()
    destination = input("Enter your travel destination: ")
    start_date = input("Enter start date (YYYY-MM-DD): ")
    end_date = input("Enter end date (YYYY-MM-DD): ")
    itinerary = asyncio.run(generate_itinerary(destination, start_date, end_date))
    print("\nGenerated Itinerary:\n", itinerary)


"""
This module contains utility functions for judge models.
"""
import litellm
from typing import Optional, Union, Tuple, List

from judgeval.common.exceptions import InvalidJudgeModelError
from judgeval.judges import JudgevalJudge, LiteLLMJudge, TogetherJudge, MixtureOfJudges
from judgeval.constants import TOGETHER_SUPPORTED_MODELS, JUDGMENT_SUPPORTED_MODELS, ACCEPTABLE_MODELS

LITELLM_SUPPORTED_MODELS = set(litellm.model_list)

def create_judge(
    model: Optional[Union[str, List[str], JudgevalJudge]] = None) -> Tuple[JudgevalJudge, bool]:
    """
    Creates a judge model from string(s) or a judgeval judge object.

    If `model` is a single string, it is assumed to be a judge model name.
    If `model` is a list of strings, it is assumed to be a list of judge model names (for MixtureOfJudges).
    If `model` is a judgeval judge object, it is returned as is.    

    Returns a tuple of (initialized judgevalBaseLLM, using_native_model boolean)
    If no model is provided, uses GPT4o as the default judge.
    """
    if model is None:  # default option
        return LiteLLMJudge(model="gpt-4o"), True
    if not isinstance(model, (str, list, JudgevalJudge)):
        raise InvalidJudgeModelError(f"Model must be a string, list of strings, or a judgeval judge object. Got: {type(model)} instead.")
    # If model is already a valid judge type, return it and mark native
    if isinstance(model, (JudgevalJudge, LiteLLMJudge, TogetherJudge, MixtureOfJudges)):
        return model, True 

    # Either string or List[str]
    if isinstance(model, list):
        for m in model:
            if m in JUDGMENT_SUPPORTED_MODELS:
                raise NotImplementedError(
                    """Judgment models are not yet supported for local scoring.
                    Please either set the `use_judgment` flag to True or use 
                    non-Judgment models."""
                )
            if m not in ACCEPTABLE_MODELS:
                raise InvalidJudgeModelError(f"Invalid judge model chosen: {m}")
        return MixtureOfJudges(models=model), True
    # If model is a string, check that it corresponds to a valid model
    if model in LITELLM_SUPPORTED_MODELS:
        return LiteLLMJudge(model=model), True
    if model in TOGETHER_SUPPORTED_MODELS:
        return TogetherJudge(model=model), True
    if model in JUDGMENT_SUPPORTED_MODELS:
        raise NotImplementedError(
            """Judgment models are not yet supported for local scoring.
            Please either set the `use_judgment` flag to True or use 
            non-Judgment models."""
        )
    else:
        raise InvalidJudgeModelError(f"Invalid judge model chosen: {model}")



"""
Implements the base class for all Judgeval Judge models.
"""

from abc import ABC, abstractmethod
from typing import Optional


class JudgevalJudge(ABC):
    def __init__(self, model_name: Optional[str] = None, *args, **kwargs):
        self.model_name = model_name
        self.model = self.load_model(*args, **kwargs)

    @abstractmethod
    def load_model(self, *args, **kwargs):
        """Loads a model, that will be responsible for scoring.

        Returns:
            A model object
        """
        pass

    @abstractmethod
    def generate(self, *args, **kwargs) -> str:
        """Runs the model to output LLM response.

        Returns:
            A string.
        """
        pass

    @abstractmethod
    async def a_generate(self, *args, **kwargs) -> str:
        """Runs the model to output LLM response.

        Returns:
            A string.
        """
        pass
    
    @abstractmethod
    def get_model_name(self, *args, **kwargs) -> str:
        pass
    


"""
Implementation of the ScorerData class.

ScorerData holds the information related to a single, completed Scorer evaluation run.
"""

from typing import List, Union, Optional, Dict
from pydantic import BaseModel

from judgeval.scorers import JudgevalScorer

class ScorerData(BaseModel):
    """
    ScorerData holds the information related to a single, completed Scorer evaluation run.

    For example, if running the Judgment Faithfulness scorer on an example, the ScorerData
    object will contain whether the example passed its threshold expectation, as well as more detailed
    information surrounding the evaluation run such as the claims and verdicts generated by the
    judge model(s).
    """
    name: str
    threshold: float
    success: bool
    score: Optional[float] = None
    reason: Optional[str] = None
    strict_mode: Optional[bool] = None
    evaluation_model: Union[List[str], str] = None
    error: Optional[str] = None
    evaluation_cost: Union[float, None] = None
    verbose_logs: Optional[str] = None
    additional_metadata: Optional[Dict] = None

    def to_dict(self) -> dict:
        """Convert the ScorerData instance to a JSON-serializable dictionary."""
        return {
            "name": self.name,
            "threshold": self.threshold,
            "success": self.success,
            "score": self.score,
            "reason": self.reason,
            "strict_mode": self.strict_mode,
            "evaluation_model": self.evaluation_model,
            "error": self.error,
            "evaluation_cost": self.evaluation_cost,
            "verbose_logs": self.verbose_logs,
            "additional_metadata": self.additional_metadata
        }


def create_scorer_data(scorer: JudgevalScorer) -> ScorerData:
    """
    After a `scorer` is run, it contains information about the example that was evaluated
    using the scorer. For example, after computing Faithfulness, the `scorer` object will contain
    whether the example passed its threshold, the score, the reason for score, etc.

    This function takes an executed `scorer` object and produces a ScorerData object that
    contains the output of the scorer run that can be exported to be logged as a part of
    the ScorerResult.
    """
    if scorer.error is not None:  # error occurred during eval run
        return ScorerData(
            name=scorer.__name__,
            threshold=scorer.threshold,
            score=None,
            reason=None,
            success=False,
            strict_mode=scorer.strict_mode,
            evaluation_model=scorer.evaluation_model,
            error=scorer.error,
            evaluation_cost=scorer.evaluation_cost,
            verbose_logs=scorer.verbose_logs,
        )
    else:  # standard execution, no error
        return ScorerData(
            name=scorer.__name__,
            score=scorer.score,
            threshold=scorer.threshold,
            reason=scorer.reason,
            success=scorer._success_check(),
            strict_mode=scorer.strict_mode,
            evaluation_model=scorer.evaluation_model,
            error=None,
            evaluation_cost=scorer.evaluation_cost,
            verbose_logs=scorer.verbose_logs,
            additional_metadata=scorer.additional_metadata,
        )



"""
Util prompts for AnswerRelevancyScorer
"""

from typing import List, Tuple
from pydantic import BaseModel


# BaseModels to enforce formatting in LLM JSON response
class Statements(BaseModel):
    statements: List[str]


class ARVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[ARVerdict]


class Reason(BaseModel):
    reason: str


class AnswerRelevancyTemplate:
    @staticmethod
    def deduce_statements(actual_output):
        return f"""You will be presented with a piece of text. Your task is to break down the text and generate a list of statements contained within the text. Single words and ambiguous phrases should be considered statements.

===== START OF EXAMPLES =====
Example 1:
Example text: The weather is sunny today. Temperature is 75 degrees. Don't forget your sunscreen!

Output:
{{
    "statements": ["The weather is sunny today", "Temperature is 75 degrees", "Don't forget your sunscreen!"]
}}

Example 2:
Example text: I love pizza. It has cheese and tomato sauce and the crust is crispy.

Output:
{{
    "statements": ["I love pizza", "It has cheese and tomato sauce", "The crust is crispy"]
}}
===== END OF EXAMPLES =====

        
**
IMPORTANT: Please return your answer in valid JSON format, with the "statements" key mapping to a list of strings. No words or explanation is needed.
**

==== START OF INPUT ====
Text:
{actual_output}
==== END OF INPUT ====

==== YOUR ANSWER ====
JSON:
"""

    @staticmethod
    def generate_verdicts(input, actual_output):
        return f"""You will be provided with a list of statements from a response; your task is to determine whether each statement is relevant with respect to a provided input.
More specifically, you should generate a JSON object with the key "verdicts". "verdicts" will map to a list of nested JSON objects with two keys: `verdict` and `reason`.
The "verdict" key be ONE OF THE FOLLOWING: ["yes", "no", "idk"]. You should select "yes" if the statement is relevant to addressing the original input, "no" if the statement is irrelevant, and 'idk' if it is ambiguous (eg., not directly relevant but could be used as a supporting point to address the input).
The "reason" key should provide an explanation for your choice, regardless of which verdict you select.

NOTE: the list of statements was generated from an output corresponding to the provided `input`. Account for this relationship during your evaluation of the content relevancy.

==== OUTPUT FORMATTING ====
IMPORTANT: Please make sure to only return in JSON format, with the "verdicts" key mapping to a list of JSON objects. Each JSON object should contain keys "verdict" (one of ["yes", "no", "idk"]) and "reason" (str). 

==== START OF EXAMPLES ====
Example input 1: How do I make chocolate chip cookies?
Example statements 1: ["Preheat the oven to 375°F.", "I love baking!", "My grandmother had a cat.", "Mix the butter and sugar until creamy.", "Have a great day!"]
Example JSON 1:
{{
    "verdicts": [
        {{
            "verdict": "yes",
            "reason": "Preheating the oven is a crucial first step in baking cookies"
        }},
        {{
            "verdict": "idk",
            "reason": "While showing enthusiasm for baking, this statement doesn't directly contribute to the recipe instructions"
        }},
        {{
            "verdict": "no",
            "reason": "The statement about the grandmother's cat is completely irrelevant to instructions for making chocolate chip cookies"
        }},
        {{
            "verdict": "yes",
            "reason": "Mixing butter and sugar is an essential step in cookie preparation"
        }},
        {{
            "verdict": "no",
            "reason": "A farewell message is not relevant to the cookie recipe instructions being requested"
        }}
    ]  
}}

Example input 2: What are the main causes of climate change?
Example statements 2: ["Greenhouse gas emissions trap heat in the atmosphere.", "I watched a movie yesterday.", "Industrial processes release large amounts of CO2.", "The weather is nice today."]
Example JSON 2:
{{
    "verdicts": [
        {{
            "verdict": "yes",
            "reason": "This directly explains a key mechanism of climate change"
        }},
        {{
            "verdict": "no",
            "reason": "Personal entertainment activities are not related to the causes of climate change"
        }},
        {{
            "verdict": "yes",
            "reason": "This identifies a major source of greenhouse gas emissions contributing to climate change"
        }},
        {{
            "verdict": "idk",
            "reason": "While weather is related to climate, a single day's weather observation doesn't directly address the causes of climate change"
        }}
    ]
}}
==== END OF EXAMPLES ====

** LASTLY **
Since you are tasked to choose a verdict for each statement, the number of "verdicts" SHOULD BE EXACTLY EQUAL to the number of "statements".


==== YOUR TURN =====

Input:
{input}

Statements:
{actual_output}

JSON:
"""

    @staticmethod
    def generate_reason(irrelevant_statements: List[Tuple[str, str]], input: str, score: float):
        irrelevant_statements = "\n".join([f"statement: {statement}\nreason: {reason}\n------" for statement, reason in irrelevant_statements])
        return f"""==== TASK INSTRUCTIONS ====\nYou will provided with three inputs: an answer relevancy score, a list of irrelevant statements made in a model's output (with the reason why it's irrelevant), and the corresponding input to the output. Your task is to provide a CLEAR and CONCISE reason for the answer relevancy score.
You should explain why the score is not higher, but also include why its current score is fair.
The irrelevant statements represent parts of the model output that are irrelevant to addressing whatever is asked/talked about in the input. The irrelevant statement will be paired with the reason why it's irrelevant.
If there are no irrelevant statements, instead respond with a positive remark with an upbeat encouraging tone (but don't overblow the kind attitude).


==== FORMATTING YOUR ANSWER ====
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The score is <answer_relevancy_score> because <your_reason>."
}}

==== YOUR TURN ====
---- ANSWER RELEVANCY SCORE ----
{score}

---- IRRELEVANT STATEMENTS ----
{irrelevant_statements}

---- INPUT ----
{input}

---- YOUR RESPONSE ----
JSON:
"""




"""
`judgeval` comparison scorer

TODO add link to docs page for this scorer

"""

# Internal imports
from judgeval.scorers.api_scorer import APIJudgmentScorer
from judgeval.constants import APIScorer
from typing import Optional, Dict
from judgeval.data import ExampleParams
class ComparisonScorer(APIJudgmentScorer):
    kwargs: Optional[Dict] = None

    def __init__(self, threshold: float, criteria: str, description: str):
        super().__init__(
            threshold=threshold, 
            score_type=APIScorer.COMPARISON,
            required_params=[
                ExampleParams.INPUT,
                ExampleParams.ACTUAL_OUTPUT,
                ExampleParams.EXPECTED_OUTPUT,
            ]
        )
        self.kwargs = {"criteria": criteria, "description": description}

    @property
    def __name__(self):
        return f"Comparison-{self.kwargs['criteria']}"
    
    def to_dict(self) -> dict:
        """
        Converts the scorer configuration to a dictionary format.
        
        Returns:
            dict: A dictionary containing the scorer's configuration
        """
        return {
            "score_type": self.score_type,
            "threshold": self.threshold,
            "kwargs": self.kwargs
        }



"""
`judgeval` faithfulness scorer

TODO add link to docs page for this scorer

"""

# Internal imports
from judgeval.scorers.api_scorer import APIJudgmentScorer
from judgeval.constants import APIScorer
from judgeval.data import ExampleParams

class FaithfulnessScorer(APIJudgmentScorer):
    def __init__(self, threshold: float):
        super().__init__(
            threshold=threshold, 
            score_type=APIScorer.FAITHFULNESS,
            required_params=[
                ExampleParams.INPUT,
                ExampleParams.ACTUAL_OUTPUT,
                ExampleParams.RETRIEVAL_CONTEXT,
            ]
        )

    @property
    def __name__(self):
        return "Faithfulness"



from typing import List, Optional
from pydantic import BaseModel, Field


class FaithfulnessVerdict(BaseModel):
    verdict: str
    reason: Optional[str] = Field(default=None)


class Verdicts(BaseModel):
    verdicts: List[FaithfulnessVerdict]


class Truths(BaseModel):
    truths: List[str]


class Claims(BaseModel):
    claims: List[str]


class Reason(BaseModel):
    reason: str


class FaithfulnessTemplate:
    @staticmethod
    def find_claims(text):
        return f"""==== TASK INSTRUCTIONS ====
You will be provided with a passage of text. Based on the text, your task is to generate a comprehensive list of ALL CLAIMS that can be inferred from the text.  
For every claim that you derive from the text, provide the source of the claim via quoting the original text. Please try to extract EVERY CLAIM that is in the original text; priortize generating the most claims rather than being concise. 
You should NOT include any prior knowledge, and take the text at face value when extracting claims.

==== FORMATTING YOUR ANSWER ====
Please return your answer in JSON format, with the "claims" key as a list of JSON objects with keys "claim" and "quote". No words or explanation beyond the output JSON is needed.

==== EXAMPLES ====

---- START OF EXAMPLE 1 ----
Example Text: 
"Einstein won the nobel prize in 1968 for his discovery of the photoelectric effect."

Example JSON: 
{{
    "claims": [
        {{
             "claim": "Einstein won the nobel prize for his discovery of the photoelectric effect.",
             "quote": "Einstein won the nobel prize in 1968 for his discovery of the photoelectric effect."
        }},
        {{
             "claim": "Einstein won the nobel prize in 1968.",
             "quote": "Einstein won the nobel prize in 1968 for his discovery of the photoelectric effect."
        }}
    ]  
}}
---- END OF EXAMPLE 1 ----

---- START OF EXAMPLE 2 ----
Example Text: "The Wright brothers successfully flew the first powered airplane on December 17, 1903, in Kitty Hawk, North Carolina."

{{
    "claims": [
        {{
            "claim": "The Wright brothers flew the first powered airplane.",
            "quote": "The Wright brothers successfully flew the first powered airplane on December 17, 1903, in Kitty Hawk, North Carolina."
        }},
        {{
            "claim": "The Wright brothers made their flight in Kitty Hawk, North Carolina.",
            "quote": "The Wright brothers successfully flew the first powered airplane on December 17, 1903, in Kitty Hawk, North Carolina."
        }},
        {{
            "claim": "The first powered airplane flight occurred on December 17, 1903.",
            "quote": "The Wright brothers successfully flew the first powered airplane on December 17, 1903, in Kitty Hawk, North Carolina."
        }}
    ]
}}
---- END OF EXAMPLE 2 ----

---- START OF EXAMPLE 3 ----
Example Text:
"The Great Wall of China was built over many centuries by different Chinese dynasties. Construction began more than 2,000 years ago during the Warring States period. The most famous sections were built during the Ming Dynasty. The wall stretches for thousands of miles across northern China and was primarily built for military defense."

Example JSON:
{{
    "claims": [
        {{
            "claim": "The Great Wall of China was built by multiple Chinese dynasties",
            "quote": "The Great Wall of China was built over many centuries by different Chinese dynasties."
        }},
        {{
            "claim": "Construction of the Great Wall began over 2,000 years ago",
            "quote": "Construction began more than 2,000 years ago during the Warring States period."
        }},
        {{
            "claim": "Construction started during the Warring States period",
            "quote": "Construction began more than 2,000 years ago during the Warring States period."
        }},
        {{
            "claim": "The most well-known parts of the wall were constructed during the Ming Dynasty",
            "quote": "The most famous sections were built during the Ming Dynasty."
        }},
        {{
            "claim": "The Great Wall extends for thousands of miles",
            "quote": "The wall stretches for thousands of miles across northern China"
        }},
        {{
            "claim": "The wall is located in northern China",
            "quote": "The wall stretches for thousands of miles across northern China"
        }},
        {{
            "claim": "The Great Wall was constructed for defensive military purposes",
            "quote": "was primarily built for military defense."
        }}
    ]
}}
---- END OF EXAMPLE 3 ----

==== END OF EXAMPLES ====

==== YOUR TURN ====

Example Text:
{text}

JSON:
"""

    @staticmethod
    def create_verdicts(claims, retrieval_context):
        return f"""==== TASK INSTRUCTIONS ====
You will be provided with a list of claims from an LLM's output text, accompanied by the retrieval documents that the LLM used to generate the output. 
I'm pretty sure that many of the claims are factually contradictory to the retrieval context, but I want you to double check that I'm right. 
For each claim, choose one of ("yes", "no", or "idk") to represent whether the claim is correct based on the retrieval context. 
YOU SHOULD be very scrutinous--if any part of the claim is contradicted by the retrieval context, you should choose "no". Think really hard about finding the contradictions, since they can be subtle!

Choose 'no' if the retrieval context CONTRADICTS the claims. YOU SHOULD NEVER USE YOUR PRIOR KNOWLEDGE IN YOUR JUDGMENT.
Claims made using vague, suggestive, or speculative language such as 'may have', 'possibility due to', do NOT count as a contradiction.
Claims that are fuzzy based on lack of information MUST BE ANSWERED with 'idk'.

==== FORMATTING YOUR ANSWER ====
Please return your answer in JSON format, with the 'verdicts' key as a list of JSON objects. Each JSON object should have 2 fields: 'verdict' and 'reason'. 
The 'verdict' key should be either 'yes', 'no', or 'idk', which states WHETHER THE CLAIM AGREES with the context. 
The 'reason' key should be a string explaining why the claim is 'yes', 'no', or 'idk'. Make specific reference to the retrieval context to justify your verdict.

==== EXAMPLES ====
---- START OF EXAMPLE 1 ----
Example retrieval contexts: "Einstein won the Nobel Prize for his discovery of the photoelectric effect. Einstein won the Nobel Prize in 1968. Einstein is a German Scientist."
Example claims: ["Barack Obama is a caucasian male.", "Zurich is a city in London", "Einstein won the Nobel Prize for the discovery of the photoelectric effect which may have contributed to his fame.", "Einstein won the Nobel Prize in 1969 for his discovery of the photoelectric effect.", "Einstein was a Germen chef."]

Example JSON:
{{
    "verdicts": [
        {{
            "verdict": "idk",
            "reason": "The claim about Barack Obama's ethnicity cannot be verified from the given retrieval context as it contains no information about Barack Obama."
        }},
        {{
            "verdict": "idk", 
            "reason": "The claim about Zurich being a city in London cannot be verified from the given retrieval context as it contains no information about Zurich or London."
        }},
        {{
            "verdict": "yes",
            "reason": "The retrieval context confirms that Einstein won the Nobel Prize for discovering the photoelectric effect."
        }},
        {{
            "verdict": "no",
            "reason": "The actual output claims Einstein won the Nobel Prize in 1969, which is untrue as the retrieval context states it is 1968 instead."
        }},
        {{
            "verdict": "no",
            "reason": "The actual output claims Einstein is a Germen chef, which is not correct as the retrieval context states he was a German scientist instead."
        }}
    ]  
}}
---- END OF EXAMPLE 1 ----
---- START OF EXAMPLE 2 ----
Example retrieval contexts: "The Great Wall of China was built over many centuries by different Chinese dynasties. Construction began more than 2,000 years ago. The wall stretches for thousands of miles across China's northern borders. Most of the existing wall was built during the Ming Dynasty."
Example claims: ["The Great Wall was built in a single year.", "The Great Wall may have taken centuries to complete.", "The Great Wall was built by the Romans.", "The Great Wall is located in China's northern region.", "The Great Wall is 100 meters long."]

Example JSON:
{{
    "verdicts": [
        {{
            "verdict": "no",
            "reason": "The claim that the Great Wall was built in a single year directly contradicts the retrieval context, which states it was built over many centuries."
        }},
        {{
            "verdict": "yes",
            "reason": "The retrieval context confirms that the Great Wall was built over many centuries by different Chinese dynasties."
        }},
        {{
            "verdict": "no",
            "reason": "The claim states the Romans built the wall, which contradicts the retrieval context that specifies it was built by Chinese dynasties."
        }},
        {{
            "verdict": "yes",
            "reason": "The retrieval context explicitly states that the wall stretches across China's northern borders."
        }},
        {{
            "verdict": "no",
            "reason": "The claim that the wall is 100 meters long contradicts the retrieval context which states it stretches for thousands of miles."
        }}
    ]
}}
---- END OF EXAMPLE 2 ----
==== END OF EXAMPLES ====

==== YOUR TURN ====
Retrieval Contexts:
{retrieval_context}

Claims:
{claims}

JSON:
"""

    @staticmethod
    def justify_reason(score, contradictions):
        return f"""==== TASK INSTRUCTIONS ====
You will be provided with a list of contradictions and a faithfulness score. 
The list of contradictions will be references to statements made by a RAG generator that contradicted one or more document(s) from the retrieval context. 
- To clarify, the LLM produced an `actual output` that contained claims that contradicted the `retrieval context` used to produce the output.
The faithfulness score is a 0 - 1 float (1 is highest) indicating how factually consistent the RAG generator's output is to the retrieval context.

Your task is to CLEARLY and CONCISELY summarize the contradictions to justify the score. 
If there are no contradictions, just say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying). 
Your reason MUST use information from the contradictions in your reason.


==== FORMATTING YOUR ANSWER ====
Please make sure to only return in JSON format, with the 'reason' key providing the reason.

Example JSON:
{{
    "reason": "The score is <faithfulness_score> because <your_reason>."
}}

==== EXAMPLE ====
Example Contradictions:
[
    {{
        "verdict": "no", 
        "reason": "The output claims Marie Curie was born in Russia, but the context clearly states she was born in Warsaw, Poland."
    }},
    {{
        "verdict": "no",
        "reason": "The output states Marie Curie discovered uranium, but the context indicates she discovered radium and polonium."
    }}
]

Example Faithfulness Score:
0.60

Example Response:
{{
    "reason": "The score is 0.60 because the output made two significant factual errors: incorrectly stating Marie Curie was born in Russia instead of Poland, and wrongly attributing the discovery of uranium to her instead of radium and polonium."
}}

==== YOUR TURN ====
Faithfulness Score:
{score}

Contradictions:
{contradictions}

JSON:
"""



from typing import List, Optional, Dict, Any, Union
from pydantic import BaseModel, ConfigDict, model_validator

from judgeval.data.example import Example
from judgeval.data.scorer_data import ScorerData
from judgeval.common.logger import debug, error

class ProcessExample(BaseModel):
    """
    ProcessExample is an `Example` object that contains intermediate information 
    about an undergoing evaluation on the original `Example`. It is used purely for
    internal operations and keeping track of the evaluation process.
    """
    name: str
    input: Optional[str] = None
    actual_output: Optional[Union[str, List[str]]] = None
    expected_output: Optional[Union[str, List[str]]] = None
    context: Optional[list] = None
    retrieval_context: Optional[list] = None
    tools_called: Optional[list] = None
    expected_tools: Optional[list] = None

    # make these optional, not all test cases in a conversation will be evaluated
    success: Optional[bool] = None
    scorers_data: Optional[List[ScorerData]] = None
    run_duration: Optional[float] = None 
    evaluation_cost: Optional[float] = None

    order: Optional[int] =  None
    # These should map 1 to 1 from golden
    additional_metadata: Optional[Dict] = None
    comments: Optional[str] = None
    trace_id: Optional[str] = None
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def update_scorer_data(self, scorer_data: ScorerData):
        """
        Updates scorer data field of test case after the scorers have been
        evaluated on this test case.
        """
        debug(f"Updating scorer data for example '{self.name}' with scorer: {scorer_data}")
        # self.scorers_data is a list of ScorerData objects that contain the 
        # evaluation results of each scorer on this test case
        if self.scorers_data is None:
            self.scorers_data = [scorer_data]
        else:
            self.scorers_data.append(scorer_data)

        if self.success is None:
            # self.success will be None when it is a message
            # in that case we will be setting success for the first time
            self.success = scorer_data.success
        else:
            if scorer_data.success is False:
                debug(f"Example '{self.name}' marked as failed due to scorer: {scorer_data}")
                self.success = False

    def update_run_duration(self, run_duration: float):
        self.run_duration = run_duration
    

def create_process_example(
    example: Example,
) -> ProcessExample:
    """
    When an LLM Test Case is executed, we track its progress using an ProcessExample.

    This will track things like the success of the test case, as well as the metadata (such as verdicts and claims in Faithfulness).
    """
    success = True
    if example.name is not None:
        name = example.name
    else:
        name = "Test Case Placeholder"
        debug(f"No name provided for example, using default name: {name}")
    order = None
    scorers_data = []

    debug(f"Creating ProcessExample for: {name}")
    process_ex = ProcessExample(
        name=name,
        input=example.input,
        actual_output=example.actual_output,
        expected_output=example.expected_output,
        context=example.context,
        retrieval_context=example.retrieval_context,
        tools_called=example.tools_called,
        expected_tools=example.expected_tools,
        success=success,
        scorers_data=scorers_data,
        run_duration=None,
        evaluation_cost=None,
        order=order,
        additional_metadata=example.additional_metadata,
        trace_id=example.trace_id
    )
    return process_ex




from judgeval.judges.base_judge import JudgevalJudge
from judgeval.judges.litellm_judge import LiteLLMJudge
from judgeval.judges.together_judge import TogetherJudge
from judgeval.judges.mixture_of_judges import MixtureOfJudges

__all__ = ["JudgevalJudge", "LiteLLMJudge", "TogetherJudge", "MixtureOfJudges"]



"""
Rules system for Judgeval that enables alerts based on metric thresholds.
"""

from typing import Dict, List, Optional, Union, Any, Set, Tuple
from pydantic import BaseModel, Field, field_validator, ConfigDict
from enum import Enum
import asyncio
from concurrent.futures import ThreadPoolExecutor
import time
import uuid

from judgeval.scorers import APIJudgmentScorer, JudgevalScorer, ScorerWrapper

class AlertStatus(str, Enum):
    """Status of an alert evaluation."""
    TRIGGERED = "triggered"
    NOT_TRIGGERED = "not_triggered"

class Condition(BaseModel):
    """
    A single metric condition.
    
    Example:
        {
            "metric": FaithfulnessScorer(threshold=0.7)  # Must be a scorer object: APIJudgmentScorer, JudgevalScorer, or ScorerWrapper
        }
        
    The Condition class uses the scorer's threshold and success function internally.
    """
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    metric: Union[APIJudgmentScorer, JudgevalScorer, ScorerWrapper]

    @property
    def metric_name(self) -> str:
        """Get the name of the metric for lookups in scores dictionary."""
        if isinstance(self.metric, ScorerWrapper):
            # Handle ScorerWrapper case specifically
            return self.metric.scorer.score_type if hasattr(self.metric.scorer, 'score_type') else str(self.metric.scorer)
        elif hasattr(self.metric, 'score_type'):
            # Handle APIJudgmentScorer and JudgevalScorer which have score_type
            return self.metric.score_type
        elif hasattr(self.metric, '__name__'):
            # Handle cases where metric has a __name__ attribute
            return self.metric.__name__
        # Fallback to string representation
        return str(self.metric)

    @property
    def threshold(self) -> float:
        """Get the threshold from the metric."""
        return self.metric.threshold if hasattr(self.metric, 'threshold') else 0.5

    def evaluate(self, value: float) -> bool:
        """
        Evaluate the condition against a value.
        Returns True if the condition passes, False otherwise.
        Uses the scorer's success check function if available.
        """
        # Store the value in the scorer
        if hasattr(self.metric, 'score'):
            self.metric.score = value
        
        # Use the scorer's success check function if available
        if hasattr(self.metric, 'success_check'):
            return self.metric.success_check()
        elif hasattr(self.metric, '_success_check'):
            return self.metric._success_check()
        else:
            # Fallback to default comparison (greater than or equal)
            return value >= self.threshold if self.threshold is not None else False

class NotificationConfig(BaseModel):
    """
    Configuration for notifications when a rule is triggered.
    
    Example:
        {
            "enabled": true,
            "communication_methods": ["email", "broadcast_slack", "broadcast_email"],
            "email_addresses": ["user1@example.com", "user2@example.com"],
            "send_at": 1632150000  # Unix timestamp (specific date/time)
        }
        
    Communication Methods:
        - "email": Send emails to specified email addresses
        - "broadcast_slack": Send broadcast notifications to all configured Slack channels
        - "broadcast_email": Send broadcast emails to all organization emails
    """
    enabled: bool = True
    communication_methods: List[str] = []
    email_addresses: Optional[List[str]] = None
    send_at: Optional[int] = None  # Unix timestamp for scheduled notifications
    
    def model_dump(self, **kwargs):
        """Convert the NotificationConfig to a dictionary for JSON serialization."""
        return {
            "enabled": self.enabled,
            "communication_methods": self.communication_methods,
            "email_addresses": self.email_addresses,
            "send_at": self.send_at
        }

class Rule(BaseModel):
    """
    Configuration for a single rule.
    
    Example:
        {
            "rule_id": "123e4567-e89b-12d3-a456-426614174000",
            "name": "Quality Check",
            "description": "Check if quality metrics meet thresholds",
            "conditions": [
                {"metric": FaithfulnessScorer(threshold=0.7)},
                {"metric": AnswerRelevancyScorer(threshold=0.8)}
            ],
            "combine_type": "all",  # "all" or "any"
            "notification": {
                "enabled": true,
                "communication_methods": ["slack", "email"],
                "email_addresses": ["user1@example.com", "user2@example.com"]
            }
        }
    """
    rule_id: str = Field(default_factory=lambda: str(uuid.uuid4()))  # Random UUID string as default value
    name: str
    description: Optional[str] = None
    conditions: List[Condition]
    combine_type: str = Field(..., pattern="^(all|any)$")  # all = AND, any = OR
    notification: Optional[NotificationConfig] = None  # Configuration for notifications
    

    def model_dump(self, **kwargs):
        """
        Custom serialization that properly handles condition serialization.
        """
        data = super().model_dump(**kwargs)
        
        # Special handling for conditions with complex metric objects
        if "conditions" in data:
            for i, condition in enumerate(data["conditions"]):
                if "metric" in condition:
                    # Get the actual metric object
                    metric_obj = self.conditions[i].metric
                    
                    # Create standardized metric representation needed by server API
                    metric_data = {
                        "score_type": "",
                        "threshold": 0.0
                    }
                    
                    # First try to use object's own serialization methods
                    if hasattr(metric_obj, "to_dict"):
                        orig_data = metric_obj.to_dict()
                        # Copy any existing fields
                        for key, value in orig_data.items():
                            metric_data[key] = value
                    elif hasattr(metric_obj, "model_dump"):
                        orig_data = metric_obj.model_dump()
                        # Copy any existing fields
                        for key, value in orig_data.items():
                            metric_data[key] = value
                    
                    # If we already have data from original serialization methods but missing required fields
                    if 'name' in metric_data and 'score_type' not in metric_data:
                        metric_data['score_type'] = metric_data['name']
                        
                    # Ensure required fields have values by checking various sources
                    if not metric_data['score_type']:
                        # Try to get score_type from different possible attributes
                        if hasattr(metric_obj, 'score_type'):
                            metric_data['score_type'] = metric_obj.score_type
                        elif hasattr(metric_obj, 'name'):
                            metric_data['score_type'] = metric_obj.name
                        else:
                            # Last resort: use string representation
                            metric_data['score_type'] = str(metric_obj)
                    
                    # Make sure threshold is set
                    if not metric_data.get('threshold') and metric_data.get('threshold') != 0.0:
                        if hasattr(metric_obj, 'threshold'):
                            metric_data['threshold'] = metric_obj.threshold
                        else:
                            # Use condition threshold if metric doesn't have one
                            metric_data['threshold'] = self.conditions[i].threshold
                    
                    # Update the condition with our properly serialized metric
                    condition["metric"] = metric_data
        
        return data

    @field_validator('conditions')
    def validate_conditions_not_empty(cls, v):
        if not v:
            raise ValueError("Conditions list cannot be empty")
        return v

    @field_validator('combine_type')
    def validate_combine_type(cls, v):
        if v not in ["all", "any"]:
            raise ValueError(f"combine_type must be 'all' or 'any', got: {v}")
        return v

class AlertResult(BaseModel):
    """
    Result of evaluating a rule.
    
    Example:
        {
            "status": "triggered",
            "rule_name": "Quality Check",
            "conditions_result": [
                {"metric": "faithfulness", "value": 0.6, "threshold": 0.7, "passed": False},
                {"metric": "relevancy", "value": 0.9, "threshold": 0.8, "passed": True}
            ],
            "rule_id": "123e4567-e89b-12d3-a456-426614174000",
            "metadata": {
                "example_id": "example_123",
                "timestamp": "20240321_123456"
            },
            "notification": {
                "enabled": true,
                "communication_methods": ["slack", "email"],
                "email_addresses": ["user1@example.com", "user2@example.com"]
            }
        }
    """
    status: AlertStatus
    rule_id: Optional[str] = None  # The unique identifier of the rule
    rule_name: str
    conditions_result: List[Dict[str, Any]]
    metadata: Dict[str, Any] = {}
    notification: Optional[NotificationConfig] = None  # Configuration for notifications
    
    @property
    def example_id(self) -> Optional[str]:
        """Get example_id from metadata for backward compatibility"""
        return self.metadata.get("example_id")
        
    @property
    def timestamp(self) -> Optional[str]:
        """Get timestamp from metadata for backward compatibility"""
        return self.metadata.get("timestamp")

class RulesEngine:
    """
    Engine for creating and evaluating rules against metrics.
    
    Example:
        ```python
        # Define rules
        rules = {
            "1": Rule(
                name="Quality Check",
                description="Check if quality metrics meet thresholds",
                conditions=[
                    Condition(metric=FaithfulnessScorer(threshold=0.7)),
                    Condition(metric=AnswerRelevancyScorer(threshold=0.8))
                ],
                combine_type="all"
            )
        }
        
        # Create rules engine
        engine = RulesEngine(rules)
        
        # Configure notifications
        engine.configure_notification(
            rule_id="1",
            enabled=True,
            communication_methods=["slack", "email"],
            email_addresses=["user@example.com"]
        )
        
        # Evaluate rules
        scores = {"faithfulness": 0.65, "relevancy": 0.85}
        results = engine.evaluate_rules(scores, {"example_id": "example_123"})
        ```
    """
    
    def __init__(self, rules: Dict[str, Rule]):
        """
        Initialize the rules engine.
        
        Args:
            rules: Dictionary mapping rule IDs to Rule objects
        """
        self.rules = rules

    def configure_notification(self, rule_id: str, enabled: bool = True, 
                              communication_methods: List[str] = None, 
                              email_addresses: List[str] = None,
                              send_at: Optional[int] = None) -> None:
        """
        Configure notification settings for a specific rule.
        
        Args:
            rule_id: ID of the rule to configure notifications for
            enabled: Whether notifications are enabled for this rule
            communication_methods: List of notification methods (e.g., ["slack", "email"])
            email_addresses: List of email addresses to send notifications to
            send_at: Optional Unix timestamp for when to send the notification
        """
        if rule_id not in self.rules:
            raise ValueError(f"Rule ID '{rule_id}' not found")
            
        rule = self.rules[rule_id]
        
        # Create notification configuration if it doesn't exist
        if rule.notification is None:
            rule.notification = NotificationConfig()
            
        # Set notification parameters
        rule.notification.enabled = enabled
        
        if communication_methods is not None:
            rule.notification.communication_methods = communication_methods
            
        if email_addresses is not None:
            rule.notification.email_addresses = email_addresses
            
        if send_at is not None:
            rule.notification.send_at = send_at
    
    def configure_all_notifications(self, enabled: bool = True, 
                                   communication_methods: List[str] = None,
                                   email_addresses: List[str] = None,
                                   send_at: Optional[int] = None) -> None:
        """
        Configure notification settings for all rules.
        
        Args:
            enabled: Whether notifications are enabled
            communication_methods: List of notification methods (e.g., ["slack", "email"])
            email_addresses: List of email addresses to send notifications to
            send_at: Optional Unix timestamp for when to send the notification
        """
        for rule_id, rule in self.rules.items():
            self.configure_notification(
                rule_id=rule_id,
                enabled=enabled,
                communication_methods=communication_methods,
                email_addresses=email_addresses,
                send_at=send_at
            )
    
    def evaluate_rules(self, scores: Dict[str, float], example_metadata: Optional[Dict[str, Any]] = None) -> Dict[str, AlertResult]:
        """
        Evaluate all rules against a set of scores.
        Returns mapping of rule IDs to their alert results.
        
        Args:
            scores: Dictionary of metric names to their score values
            example_metadata: Optional dictionary containing example metadata (example_id, timestamp)
        """
        results = {}

        for rule_id, rule in self.rules.items():
            # Evaluate each condition
            condition_results = []
            passed_conditions = []
            
            for condition in rule.conditions:
                # Get the metric name for lookup
                metric_name = condition.metric_name
                value = scores.get(metric_name)
                
                if value is None:
                    # Skip this condition instead of evaluating it as false
                    condition_results.append({
                        "metric": metric_name,
                        "value": None,
                        "threshold": condition.threshold,
                        "passed": None,  # Using None to indicate the condition was skipped
                        "skipped": True  # Add a flag to indicate this condition was skipped
                    })
                    continue  # Skip adding to passed_conditions
                else:
                    passed = condition.evaluate(value)
                    condition_results.append({
                        "metric": metric_name,
                        "value": value,
                        "threshold": condition.threshold,
                        "passed": passed,
                        "skipped": False  # Indicate this condition was evaluated
                    })
                    passed_conditions.append(passed)
            
            # Determine if alert should trigger - only consider conditions that weren't skipped
            if not passed_conditions:
                # If all conditions were skipped, the rule doesn't trigger
                triggered = False
            else:
                if rule.combine_type == "all":
                    # For "all" combine_type:
                    # - All evaluated conditions must pass
                    # - All conditions must have been evaluated (none skipped)
                    all_conditions_passed = all(passed_conditions)
                    all_conditions_evaluated = len(passed_conditions) == len(rule.conditions)
                    triggered = all_conditions_passed and all_conditions_evaluated
                else:
                    # For "any" combine_type, at least one condition must pass
                    triggered = any(passed_conditions)
            
            # Create alert result with example metadata
            notification_config = None
            if triggered and rule.notification:
                # If rule has a notification config and the alert is triggered, include it in the result
                notification_config = rule.notification
            
            # Set the alert status based on whether the rule was triggered
            status = AlertStatus.TRIGGERED if triggered else AlertStatus.NOT_TRIGGERED
            
            # Create the alert result
            alert_result = AlertResult(
                status=status,
                rule_id=rule.rule_id,
                rule_name=rule.name,
                conditions_result=condition_results,
                notification=notification_config,
                metadata=example_metadata or {}
            )
            
            results[rule_id] = alert_result
            
        return results
    
    async def evaluate_rules_parallel(self, 
                               example_scores: Dict[str, Dict[str, float]], 
                               example_metadata: Dict[str, Dict[str, Any]],
                               max_concurrent: int = 100) -> Dict[str, Dict[str, AlertResult]]:
        """
        Evaluate all rules against multiple examples in parallel.
        
        Args:
            example_scores: Dictionary mapping example_ids to their score dictionaries
            example_metadata: Dictionary mapping example_ids to their metadata
            max_concurrent: Maximum number of concurrent evaluations
            
        Returns:
            Dictionary mapping example_ids to dictionaries of rule_ids and their alert results
        """
        # Create semaphore to limit concurrent executions
        semaphore = asyncio.Semaphore(max_concurrent)
        results = {}
        tasks = []
        
        # Create a task for each example
        for example_id, scores in example_scores.items():
            metadata = example_metadata.get(example_id, {})
            task = self._evaluate_with_semaphore(
                semaphore=semaphore,
                example_id=example_id,
                scores=scores,
                metadata=metadata
            )
            tasks.append(task)
        
        # Run all tasks and collect results
        example_results = await asyncio.gather(*tasks)
        
        # Organize results by example_id
        for example_id, result in example_results:
            results[example_id] = result
            
        return results
    
    async def _evaluate_with_semaphore(self, 
                                semaphore: asyncio.Semaphore, 
                                example_id: str, 
                                scores: Dict[str, float], 
                                metadata: Dict[str, Any]) -> Tuple[str, Dict[str, AlertResult]]:
        """
        Helper method to evaluate rules for an example with semaphore control.
        
        Args:
            semaphore: Semaphore to control concurrency
            example_id: ID of the example being evaluated
            scores: Dictionary of scores for this example
            metadata: Metadata for this example
            
        Returns:
            Tuple of (example_id, rule_results)
        """
        async with semaphore:
            # Run the evaluation in a thread pool to avoid blocking the event loop
            # for CPU-bound operations
            with ThreadPoolExecutor() as executor:
                start_time = time.perf_counter()
                rule_results = await asyncio.get_event_loop().run_in_executor(
                    executor, 
                    self.evaluate_rules,
                    scores,
                    metadata
                )
                end_time = time.perf_counter()
                
                return (example_id, rule_results) 


"""
`judgeval` tool correctness scorer

TODO add link to docs page for this scorer

"""

# Internal imports
from judgeval.scorers.api_scorer import APIJudgmentScorer
from judgeval.constants import APIScorer
from typing import Optional, Dict, List
from judgeval.data import ExampleParams

class ExecutionOrderScorer(APIJudgmentScorer):
    kwargs: Optional[Dict] = None

    def __init__(self, threshold: float, should_exact_match: bool = False, should_consider_ordering: bool = False):
        super().__init__(
            threshold=threshold, 
            score_type=APIScorer.EXECUTION_ORDER,
            required_params=[
                ExampleParams.ACTUAL_OUTPUT,
                ExampleParams.EXPECTED_OUTPUT,
            ]
        )
        self.kwargs = {"should_exact_match": should_exact_match, "should_consider_ordering": should_consider_ordering}

    @property
    def __name__(self):
        return "Execution Order"

    def to_dict(self) -> dict:
        """
        Converts the scorer configuration to a dictionary format.
        
        Returns:
            dict: A dictionary containing the scorer's configuration
        """
        return {
            "score_type": self.score_type,
            "threshold": self.threshold,
            "kwargs": self.kwargs
        }


from typing import List, Optional, Dict, Any, Union
from pydantic import BaseModel, field_validator

from judgeval.data import Example
from judgeval.scorers import JudgevalScorer, APIJudgmentScorer
from judgeval.constants import ACCEPTABLE_MODELS
from judgeval.common.logger import debug, error
from judgeval.judges import JudgevalJudge
from judgeval.rules import Rule

class EvaluationRun(BaseModel):
    """
    Stores example and evaluation scorers together for running an eval task
    
    Args: 
        project_name (str): The name of the project the evaluation results belong to
        eval_name (str): A name for this evaluation run
        examples (List[Example]): The examples to evaluate
        scorers (List[Union[JudgmentScorer, JudgevalScorer]]): A list of scorers to use for evaluation
        model (str): The model used as a judge when using LLM as a Judge
        aggregator (Optional[str]): The aggregator to use for evaluation if using Mixture of Judges
        metadata (Optional[Dict[str, Any]]): Additional metadata to include for this evaluation run, e.g. comments, dataset name, purpose, etc.
        judgment_api_key (Optional[str]): The API key for running evaluations on the Judgment API
        rules (Optional[List[Rule]]): Rules to evaluate against scoring results
    """

    # The user will specify whether they want log_results when they call run_eval
    log_results: bool = False  # NOTE: log_results has to be set first because it is used to validate project_name and eval_name
    organization_id: Optional[str] = None
    project_name: Optional[str] = None
    eval_name: Optional[str] = None
    examples: List[Example]
    scorers: List[Union[APIJudgmentScorer, JudgevalScorer]]
    model: Union[str, List[str], JudgevalJudge]
    aggregator: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    # API Key will be "" until user calls client.run_eval(), then API Key will be set
    judgment_api_key: Optional[str] = ""
    override: Optional[bool] = False
    rules: Optional[List[Rule]] = None
    
    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)

        data["scorers"] = [
            scorer.to_dict() if hasattr(scorer, "to_dict")
            else scorer.model_dump() if hasattr(scorer, "model_dump")
            else {"score_type": scorer.score_type, "threshold": scorer.threshold}
            for scorer in self.scorers
        ]

        if self.rules:
            # Process rules to ensure proper serialization
              data["rules"] = [rule.model_dump() for rule in self.rules]
            
        return data

    @field_validator('log_results', mode='before')
    def validate_log_results(cls, v):
        if not isinstance(v, bool):
            raise ValueError(f"log_results must be a boolean. Received {v} of type {type(v)}")
        return v

    @field_validator('project_name')
    def validate_project_name(cls, v, values):
        if values.data.get('log_results', False) and not v:
            debug("No project name provided when log_results is True")
            error("Validation failed: Project name required when logging results")
            raise ValueError("Project name is required when log_results is True. Please include the project_name argument.")
        return v

    @field_validator('eval_name')
    def validate_eval_name(cls, v, values):
        if values.data.get('log_results', False) and not v:
            debug("No eval name provided when log_results is True") 
            error("Validation failed: Eval name required when logging results")
            raise ValueError("Eval name is required when log_results is True. Please include the eval_run_name argument.")
        return v

    @field_validator('examples')
    def validate_examples(cls, v):
        if not v:
            raise ValueError("Examples cannot be empty.")
        for ex in v:
            if not isinstance(ex, Example):
                raise ValueError(f"Invalid type for Example: {type(ex)}")
        return v

    @field_validator('scorers')
    def validate_scorers(cls, v):
        if not v:
            raise ValueError("Scorers cannot be empty.")
        for s in v:
            if not isinstance(s, APIJudgmentScorer) and not isinstance(s, JudgevalScorer):
                raise ValueError(f"Invalid type for Scorer: {type(s)}")
        return v

    @field_validator('model')
    def validate_model(cls, v, values):
        if not v:
            raise ValueError("Model cannot be empty.")
        
        # Check if model is a judgevalJudge
        if isinstance(v, JudgevalJudge):
            # Verify all scorers are JudgevalScorer when using judgevalJudge
            scorers = values.data.get('scorers', [])
            if not all(isinstance(s, JudgevalScorer) for s in scorers):
                raise ValueError("When using a judgevalJudge model, all scorers must be JudgevalScorer type")
            return v
            
        # Check if model is string or list of strings
        if isinstance(v, str):
            if v not in ACCEPTABLE_MODELS:
                raise ValueError(f"Model name {v} not recognized. Please select a valid model name.)")
            return v
            
        if isinstance(v, list):
            if not all(isinstance(m, str) for m in v):
                raise ValueError("When providing a list of models, all elements must be strings")
            for m in v:
                if m not in ACCEPTABLE_MODELS:
                    raise ValueError(f"Model name {m} not recognized. Please select a valid model name.")
            return v
        raise ValueError(f"Model must be one of: string, list of strings, or JudgevalJudge instance. Received type {type(v)}.")

    @field_validator('aggregator', mode='before')
    def validate_aggregator(cls, v, values):
        model = values.data.get('model')
        if isinstance(model, list) and v is None:
            raise ValueError("Aggregator cannot be empty.")
            
        if isinstance(model, list) and not isinstance(v, str):
            raise ValueError("Aggregator must be a string if provided.")
            
        if v is not None and v not in ACCEPTABLE_MODELS:
            raise ValueError(f"Model name {v} not recognized.")
            
        return v
    
    class Config:
        arbitrary_types_allowed = True



from typing import Type, Optional, Any

# Import implementations
from judgeval.scorers.judgeval_scorers.api_scorers import (
    ExecutionOrderScorer as APIExecutionOrderScorer,
    JSONCorrectnessScorer as APIJSONCorrectnessScorer,
    SummarizationScorer as APISummarizationScorer,
    HallucinationScorer as APIHallucinationScorer,
    FaithfulnessScorer as APIFaithfulnessScorer,
    ContextualRelevancyScorer as APIContextualRelevancyScorer,
    ContextualPrecisionScorer as APIContextualPrecisionScorer,
    ContextualRecallScorer as APIContextualRecallScorer,
    AnswerRelevancyScorer as APIAnswerRelevancyScorer,
    AnswerCorrectnessScorer as APIAnswerCorrectnessScorer,  
    ComparisonScorer as APIComparisonScorer,
    InstructionAdherenceScorer as APIInstructionAdherenceScorer,
    GroundednessScorer as APIGroundednessScorer,
)

from judgeval.scorers.judgeval_scorers.local_implementations import (
    AnswerRelevancyScorer as LocalAnswerRelevancyScorer,
    ContextualPrecisionScorer as LocalContextualPrecisionScorer,
    ContextualRecallScorer as LocalContextualRecallScorer,
    ContextualRelevancyScorer as LocalContextualRelevancyScorer,
    FaithfulnessScorer as LocalFaithfulnessScorer,
    JsonCorrectnessScorer as LocalJsonCorrectnessScorer,
    ExecutionOrderScorer as LocalExecutionOrderScorer,
    HallucinationScorer as LocalHallucinationScorer,
    SummarizationScorer as LocalSummarizationScorer,
    AnswerCorrectnessScorer as LocalAnswerCorrectnessScorer,
    ComparisonScorer as LocalComparisonScorer,
    InstructionAdherenceScorer as LocalInstructionAdherenceScorer,
)

from judgeval.scorers.judgeval_scorers.classifiers import Text2SQLScorer


class ScorerWrapper:
    """
    Wrapper class that can dynamically load either API or local implementation of a scorer.
    """
    def __init__(self, api_implementation: Type, local_implementation: Optional[Type] = None):
        self.api_implementation = api_implementation
        self.local_implementation = local_implementation
        self._instance = None
        self._init_args = None
        self._init_kwargs = None
        
    def __call__(self, *args, **kwargs):
        """Store initialization arguments for later use when implementation is loaded"""
        self._init_args = args
        self._init_kwargs = kwargs
        return self
    
    def load_implementation(self, use_judgment: bool = True) -> Any:
        """
        Load the appropriate implementation based on the use_judgment flag.
        
        Args:
            use_judgment (bool): If True, use API implementation. If False, use local implementation.
        
        Returns:
            Instance of the appropriate implementation
        
        Raises:
            ValueError: If local implementation is requested but not available
        """
        if self._instance is not None:
            return self._instance
            
        if use_judgment:
            implementation = self.api_implementation
        else:
            if self.local_implementation is None:
                raise ValueError("No local implementation available for this scorer")
            implementation = self.local_implementation
            
        args = self._init_args or ()
        kwargs = self._init_kwargs or {}
        self._instance = implementation(*args, **kwargs)
        return self._instance
    
    def __getattr__(self, name):
        """Defer all attribute access to the loaded implementation"""
        if self._instance is None:
            raise RuntimeError("Implementation not loaded. Call load_implementation() first")
        return getattr(self._instance, name)

# Create wrapped versions of all scorers

AnswerCorrectnessScorer = ScorerWrapper(
    api_implementation=APIAnswerCorrectnessScorer,
    local_implementation=LocalAnswerCorrectnessScorer
)

AnswerRelevancyScorer = ScorerWrapper(
    api_implementation=APIAnswerRelevancyScorer,
    local_implementation=LocalAnswerRelevancyScorer
)

ExecutionOrderScorer = ScorerWrapper(
    api_implementation=APIExecutionOrderScorer,
    local_implementation=LocalExecutionOrderScorer
)

JSONCorrectnessScorer = ScorerWrapper(
    api_implementation=APIJSONCorrectnessScorer,
    local_implementation=LocalJsonCorrectnessScorer
)

SummarizationScorer = ScorerWrapper(
    api_implementation=APISummarizationScorer,
    local_implementation=LocalSummarizationScorer
)

HallucinationScorer = ScorerWrapper(
    api_implementation=APIHallucinationScorer,
    local_implementation=LocalHallucinationScorer
)

FaithfulnessScorer = ScorerWrapper(
    api_implementation=APIFaithfulnessScorer,
    local_implementation=LocalFaithfulnessScorer
)

ContextualRelevancyScorer = ScorerWrapper(
    api_implementation=APIContextualRelevancyScorer,
    local_implementation=LocalContextualRelevancyScorer
)

ContextualPrecisionScorer = ScorerWrapper(
    api_implementation=APIContextualPrecisionScorer,
    local_implementation=LocalContextualPrecisionScorer
)

ContextualRecallScorer = ScorerWrapper(
    api_implementation=APIContextualRecallScorer,
    local_implementation=LocalContextualRecallScorer
)

InstructionAdherenceScorer = ScorerWrapper(
    api_implementation=APIInstructionAdherenceScorer,
    local_implementation=LocalInstructionAdherenceScorer
)

def ComparisonScorer(threshold: float, criteria: str, description: str):
    return ScorerWrapper(
        api_implementation=APIComparisonScorer,
        local_implementation=LocalComparisonScorer
    )(threshold=threshold, criteria=criteria, description=description)

GroundednessScorer = ScorerWrapper(
    api_implementation=APIGroundednessScorer,
)

__all__ = [
    "ExecutionOrderScorer",
    "JSONCorrectnessScorer",
    "SummarizationScorer",
    "HallucinationScorer",
    "FaithfulnessScorer",
    "ContextualRelevancyScorer",
    "ContextualPrecisionScorer",
    "ContextualRecallScorer",
    "AnswerRelevancyScorer",
    "Text2SQLScorer",
    "ComparisonScorer",
    "GroundednessScorer",
]



"""
Implementation for Mixture of Judges model through Judgeval

Enables client to use multiple models to generate responses and then aggregate them into a single response.
"""
from judgeval import *
import pydantic
from typing import List, Union, Mapping
from judgeval.judges import JudgevalJudge
from judgeval.common.utils import (
    get_completion_multiple_models, 
    get_chat_completion, 
    aget_completion_multiple_models, 
    aget_chat_completion
)
from judgeval.common.logger import debug, error

def build_dynamic_mixture_prompt(
        judge_responses: List[str], 
        custom_system_prompt: str = None, 
        custom_conversation_history: List[Mapping] = None
    ) -> List[Mapping]:
    """
    Dynamically builds a prompt to mix judge responses together for the Mixture of Judges model.

    In this implementation, we simply concatenate the judge responses into a formatted string, then
    pass it into a default prompt template. This template can be customized by providing a custom prompt.

    Args:
        judge_responses (List[str]): List of responses from individual judges to be synthesized
        custom_system_prompt (str, optional): Custom system prompt to override the default one. Defaults to None.
        custom_conversation_history (List[Mapping], optional): Custom conversation history to override the default one. Defaults to None.
    """
    formatted_responses = "\n".join([f"# Judge {i + 1}'s response: #\n{response}" for i, response in enumerate(judge_responses)])
    
    # This is the default prompt for the Mixture of Judges model
    """
    You are tasked with synthesizing responses from multiple expert judges. You will receive N individual answers on the same topic. Your job is to:

    1. Analyze and compare the key points, patterns, and agreements between the answers.
    2. Identify the consensus by focusing on areas where most or all of the answers align. Consider common reasoning and frequently mentioned conclusions.
    3. Condense the responses into a single, coherent, and concise answer that represents the collective judgment of the group.
    4. When opinions differ or contradict, highlight the most supported viewpoint while briefly acknowledging the dissenting perspectives.
    5. Ensure the final answer is balanced and clear, providing a comprehensive summary that captures the wisdom of all judges while avoiding repetition.

    ## Start of Judge Responses ##
    {{judge_responses}}
    ## End of Judge Responses ##
    Synthesized response:
    """

    default_conversation = [  # inject the judge responses into the default prompt
        {
            'role': 'system',
            'content': 'You are tasked with synthesizing responses from multiple expert judges. You will receive N individual answers on the same topic. Your job is to:\n1. Analyze and compare the key points, patterns, and agreements between the answers.\n2. Identify the consensus by focusing on areas where most or all of the answers align. Consider common reasoning and frequently mentioned conclusions.\n3. Condense the responses into a single, coherent, and concise answer that represents the collective judgment of the group.\n4. When opinions differ or contradict, highlight the most supported viewpoint while briefly acknowledging the dissenting perspectives.\n5. Ensure the final answer is balanced and clear, providing a comprehensive summary that captures the wisdom of all judges while avoiding repetition.\n\n**IMPORTANT**: IF THE JUDGE RESPONSES ARE IN JSON FORMAT, YOU MUST RESPOND USING THE SAME JSON FORMAT THAT THE RESPONSES ARE IN. If the judge responses are in JSON, you MUST RESPOND IN VALID JSON FORMAT. '
        }, 
        {
            'role': 'user',
            'content': '## Start of Judge Responses ## \n# Judge 1\'s response: #\n{\n"claims": [\n{\n"claim": "A 30-day full refund is offered.",\n"quote": "We offer a 30-day full refund at no extra cost."\n},\n{\n"claim": "The 30-day full refund comes at no extra cost.",\n"quote": "We offer a 30-day full refund at no extra cost."\n}\n]\n}\n\n# Judge 2\'s response: #\n{\n    "claims": [\n        {\n            "claim": "A full refund is offered within 30 days.",\n            "quote": "We offer a 30-day full refund at no extra cost."\n        },\n        {\n            "claim": "The 30-day full refund is offered at no extra cost.",\n            "quote": "We offer a 30-day full refund at no extra cost."\n        }\n    ]\n}\n# Judge 3\'s response: #\n {\n    "claims": [\n        {\n            "claim": "A 30-day full refund is offered.",\n            "quote": "We offer a 30-day full refund at no extra cost."\n        },\n        {\n            "claim": "The 30-day full refund is offered at no extra cost.",\n            "quote": "We offer a 30-day full refund at no extra cost."\n        }\n    ]\n}\n## End of Judge Responses ##\nSynthesized response:'
        }, 
        {
            'role': 'assistant',
            'content': 'The consensus among the judges is clear and unanimous. All three judges agree that a 30-day full refund is offered, and this refund is available at no extra cost. This conclusion is consistently supported by their statements, as each of their claims is directly quoted as: "We offer a 30-day full refund at no extra cost." There are no dissenting perspectives or opposing views provided in any of the responses, indicating complete alignment on this topic.\n\nJSON:\n{\n    "claims": [\n        {\n            "claim": "A full refund is offered within 30 days.",\n            "quote": "We offer a 30-day full refund at no extra cost."\n        },\n        {\n            "claim": "The 30-day full refund is offered at no extra cost.",\n            "quote": "We offer a 30-day full refund at no extra cost."\n        }\n    ]\n}'
        },
        {
            'role': 'user',
            'content': "## Start of Judge Responses ##\n# Judge 1's response: # \nThe capital of France is Paris.\n\n# Judge 2's response: #\nThe capital of France is Paris.\n\n# Judge 3's response: # \nThe capital of France is Paris. It's one of the most popular tourist destinations in the world, known for its art, culture, and history. It's also famous for its iconic landmarks such as the Eiffel Tower, Louvre Museum, and Notre-Dame Cathedral.\n\n## End of Judge Responses ##\nSynthesized response:"
        },
        {
            'role': 'assistant',
            'content': "The capital of France is Paris. It is widely recognized as one of the world's most popular tourist destinations, celebrated for its rich art, culture, and history. Paris is renowned for its iconic landmarks, including the Eiffel Tower, Louvre Museum, and Notre-Dame Cathedral."
        }, 
        {
            'role': 'user',
            'content': f'## Start of Judge Responses ##\n{formatted_responses}\n## End of Judge Responses ##\nSynthesized response:\n'
        }
    ]
    
    # If a custom system prompt is provided, validate and use it
    if custom_system_prompt is not None:
        if not isinstance(custom_system_prompt, str):
            error(f"TypeError: Custom system prompt must be a string. Received: {type(custom_system_prompt)}.")
            raise TypeError(f"Custom system prompt must be a string. Received: {type(custom_system_prompt)}.")
        if not custom_system_prompt:
            error("ValueError: Custom system prompt cannot be empty")
            raise ValueError("Custom system prompt cannot be empty")
        # Override the default system prompt, but also add special instructions for handling JSON
        default_conversation[0]['content'] = custom_system_prompt + "\n\n**IMPORTANT**: IF THE JUDGE RESPONSES ARE IN JSON FORMAT, YOU MUST RESPOND USING THE SAME JSON FORMAT THAT THE RESPONSES ARE IN. If the judge responses are in JSON, you MUST RESPOND IN VALID JSON FORMAT."
    
    # If a custom conversation history is provided, append the judge responses to it
    if custom_conversation_history is not None:
        # Validate custom conversation history format
        for message in custom_conversation_history:
            if not isinstance(message, dict):
                error(f"TypeError: Custom conversation history must be a list of dictionaries. Received: {message}.")
                raise TypeError(f"Custom conversation history must be a list of dictionaries. Received: {message}.")
            
            if 'role' not in message or 'content' not in message:
                error("ValueError: Each message must have 'role' and 'content' keys")
                raise ValueError("Each message must have 'role' and 'content' keys")
            
            if not isinstance(message['role'], str) or not isinstance(message['content'], str):
                error(f"TypeError: Message role and content must be strings. Received: {type(message['role'])}, {type(message['content'])}.")
                raise TypeError(f"Message role and content must be strings. Received: {type(message['role'])}, {type(message['content'])}.")
            
            if message['role'] not in ['system', 'user', 'assistant']:
                error(f"ValueError: Message role must be one of: 'system', 'user', 'assistant'. Received: {message['role']}.")
                raise ValueError(f"Message role must be one of: 'system', 'user', 'assistant'. Received: {message['role']}.")
            
        judge_responses_prompt = {
            'role': 'user', 
            'content': f'## Start of Judge Responses ##\n{formatted_responses}\n## End of Judge Responses ##\nSynthesized response:\n'
        }
        return custom_conversation_history + [judge_responses_prompt]
    
    # Otherwise return the default conversation with system prompt and examples
    # No customization, return the default conversation with system prompt and examples
    return default_conversation

BASE_CONVERSATION = [
    {"role": "system", "content": "You are a helpful assistant."},
]  # for string inputs, we need to add the user query to a base conversation, since LiteLLM only accepts a list of dictionaries as a chat history
class MixtureOfJudges(JudgevalJudge):
    """
    IMPORTANT: When supplying custom prompts and conversation histories for aggregation, supply them in the following format:
    in kwargs:
    {
        "custom_prompt": "Your custom prompt here",
        "custom_conversation": [
            {"role": "system", "content": "System message 1"},
            {"role": "user", "content": "User message 1"},
            {"role": "assistant", "content": "Assistant message 1"},
            ...
        ]
    }
    """
    def __init__(self, 
                 models: List[str] = ['QWEN', 'LLAMA3_70B_INSTRUCT_TURBO', 'MISTRAL_8x22B_INSTRUCT'],
                 aggregator: str = 'gpt-4o', 
                 **kwargs):
        """
        `models` are the individual judge models to be used for generating responses.
        `aggregator` is the model that will aggregate the responses from the individual judges.

        kwargs include "custom_prompt" and "custom_conversation" for customizing the prompt for the Mixture of Judges model.
        """
        self.models = models
        self.aggregator = aggregator
        self.kwargs = kwargs
        super().__init__(model_name=models)

    def generate(
            self, 
            input: Union[str, List[Mapping[str, str]]], 
            response_schema: pydantic.BaseModel = None, 
            aggregation_schema: pydantic.BaseModel = None,
            **kwargs) -> str:
        """
        Args:
            input (Union[str, List[Mapping[str, str]]]): Input query or conversation history to the model.
            response_schema (pydantic.BaseModel): Response schema for individual judge models.
            aggregation_schema (pydantic.BaseModel): Response schema for the aggregator model.
            kwargs: Additional keyword arguments.
        """
        debug(f"Generating response for input type: {type(input)}")
        
        # Convert input to conversation format if needed
        if isinstance(input, str):
            convo = BASE_CONVERSATION + [{"role": "user", "content": input}]
        elif isinstance(input, list):
            convo = input
        else:
            error(f"Invalid input type received: {type(input)}")
            raise TypeError(f"Input must be a string or a list of dictionaries. Input type of: {type(input)}")

        try:
            responses = get_completion_multiple_models(
                models=self.models,
                messages=[convo] * len(self.models),
                response_formats=[response_schema] * len(self.models)
            )
        except Exception as e:
            error(f"Error getting completions from multiple models: {str(e)}")
            raise

        compiled_mixture_prompt = build_dynamic_mixture_prompt(responses, self.kwargs.get('custom_prompt'), self.kwargs.get('custom_conversation'))
        
        try:
            mixed_response = get_chat_completion(
                model_type=self.aggregator,
                messages=compiled_mixture_prompt,
                response_format=aggregation_schema,
            )
        except Exception as e:
            error(f"Error getting chat completion from aggregator: {str(e)}")
            raise
            
        return mixed_response

    async def a_generate(
            self, 
            input: Union[str, List[Mapping[str, str]]], 
            response_schema: pydantic.BaseModel = None,
            aggregation_schema: pydantic.BaseModel = None,
            **kwargs
        ) -> str:
        """
        Args:
            input (Union[str, List[Mapping[str, str]]]): Input query or conversation history to the model.
            response_schema (pydantic.BaseModel): Response schema for individual judge models.
            aggregation_schema (pydantic.BaseModel): Response schema for the aggregator model.
            kwargs: Additional keyword arguments.
        """
        debug(f"Generating response for input type: {type(input)}")
        
        # Convert input to conversation format if needed
        if isinstance(input, str):
            convo = BASE_CONVERSATION + [{"role": "user", "content": input}]
        elif isinstance(input, list):
            convo = input
        else:
            error(f"Invalid input type received: {type(input)}")
            raise TypeError(f"Input must be a string or a list of dictionaries. Input type of: {type(input)}")

        try:
            responses = await aget_completion_multiple_models(
                models=self.models,
                messages=[convo] * len(self.models),
                response_formats=[response_schema] * len(self.models)
            )
        except Exception as e:
            error(f"Error getting async completions from multiple models: {str(e)}")
            raise

        compiled_mixture_prompt = build_dynamic_mixture_prompt(responses, self.kwargs.get('custom_prompt'), self.kwargs.get('custom_conversation'))
        
        try:
            mixed_response = await aget_chat_completion(
                model_type=self.aggregator,
                messages=compiled_mixture_prompt,
                response_format=aggregation_schema,
            )
        except Exception as e:
            error(f"Error getting async chat completion from aggregator: {str(e)}")
            raise
            
        return mixed_response
    
    def load_model(self):
        return self.models

    def get_model_name(self) -> List[str]:
        return self.models
